local format = string.format
local find, sub = string.find, string.sub
local concat = table.concat
local function split(str, by)
  local by_start, by_end= find(str, by, nil, true)
  if by_start then
    return sub(str, 1, by_start -1 ), sub(str, by_end + 1)
  end
  return str
end

pg_info = pg_info or {}

do
  local column_type_q = spi.prepare([[
      select quote_ident(udt_schema::text) || '.' || quote_ident(udt_name::text) as tp
      from information_schema.columns
      where table_schema = $1 and table_name = $2 and column_name = $3 ]],
      'information_schema.sql_identifier','information_schema.sql_identifier','information_schema.sql_identifier')
  function pg_info.get_column_type(schema, table, name)
      local res, str=column_type_q:execute(schema, table, name)
      --spi.notice(tostring(res)..'|'..tostring(str))
      if not res or #res == 0  then
        spi.error(("Column %q does not exists for table %q.%q"):format(name, schema, table))
      end
      return res[1].tp
  end
  local visible_table_name_q = spi.prepare([[
        SELECT nspname, relname
        FROM pg_class 
        join pg_namespace 
          on pg_namespace.oid = relnamespace
        WHERE pg_table_is_visible(pg_class.oid)
          and relname = $1]], 'name')
  function pg_info.visible_table_schema_name(name)
      local res = visible_table_name_q(name)
      if spi.empty(res) then
        spi.error("There is no visible table "..name)
      end
      return res[1].nspname, res[1].relname
  end
end

pg_table = pg_table or {}

do

    local pg_table_object_funcs = pg_table.funcs or {}
    pg_table.funcs = pg_table_object_funcs
    pg_table.__cache = {}
    local pg_table_object_mt = {__index = pg_table_object_funcs}
    local function new_table_object(pg_table, tabname)
      local schema, tablen = split(tabname, '.')
      if not tablen then
        schema, tablen = pg_info.visible_table_schema_name(tabname)
        tabname = spi.quote_ident(schema) .. '.' .. spi.quote_ident(tablen)
      end
      local obj = pg_table.__cache[tabname]
      if not obj then
        obj = setmetatable({schema = schema, table = tablen, quoted = tabname }, pg_table_object_mt)
        pg_table.__cache[tabname] = obj
      end
      return obj
    end
    pg_table_mt = {__index = new_table_object}
    setmetatable(pg_table, pg_table_mt)



    local function quote_fields(fields)
        return { spi.quote_ident( unpack(fields) ) }
    end
    local function fields_equiv(fields, start_num)
        local res, start_num = {}, start_num or 1
        local quoted = quote_fields(fields)
        for i = 1, #quoted do
          res[i] = format("%s = $%d", quoted[i], start_num)
          start_num = start_num + 1
        end
        return res, start_num
    end
    local function fields_select(fields, default)
      local select
      if fields then
        select = fields.raw and fields or quote_fields(fields)
        select = concat(select, ', ')
      else
        select = default
      end
      return select
    end
    
    local function row_finder(prepared)
      return function(...)
        local res = prepared:execute( spi.limit(1), ... )
        if spi.empty(res) then
          return nil
        end
        return res[1]
      end
    end
    
    function pg_table_object_funcs:field_types( fields )
        local res = {}
        for i = 1, #fields do
          res[i] = pg_info.get_column_type(self.schema, self.table, fields[i])
        end
        return res
    end
    
    function pg_table_object_funcs:finder( fields, fields_to_select )
      local select_fields = fields_select(fields_to_select, '*')
      local conditions_sql = fields_equiv(fields)
      conditions_sql = concat(conditions_sql, "\n   AND ")
      local conditions_types = self:field_types(fields)
      local query = format("select %s\nfrom %s\nwhere %s", select_fields, self.quoted, conditions_sql)
      return spi.prepare(query, unpack(conditions_types))
    end
    
    function pg_table_object_funcs:row_finder( fields, fields_to_select )
      return row_finder(self:finder( fields, fields_to_select ))
    end
    
    function pg_table_object_funcs:row_inserter( fields, returning )
      returning = fields_select(returning, '')
      if returning ~= '' then
        returning = 'returning '..returning
      end
      local quoted_fields = concat( quote_fields(fields), ', ' )
      local field_types = self:field_types(fields)
      local field_numbers = {}
      for i=1, #fields do
        field_numbers[i] = format("$%d", i)
      end
      field_numbers = concat(field_numbers, ', ')
      local query = format("insert into %s (%s) values (%s) %s", self.quoted, quoted_fields, field_numbers, returning)
      local prepared = spi.prepare(query, unpack(field_types))
      return returning == '' and prepared or row_finder(prepared)
    end
    
    function pg_table_object_funcs:row_updater( fields_conditions, fields_to_update, returning)
      returning = fields_select(returning, '')
      if returning ~= '' then
        returning = 'returning '..returning
      end
      local conditions, next_num = fields_equiv(fields_conditions)
      local updates = fields_equiv(fields_to_update, next_num)
      updates = concat(updates, ",\n    ")
      conditions = concat(conditions, "\n   AND ")
      local types = self:field_types(fields_conditions)
      local update_types = self:field_types(fields_to_update)
      for i = 1, #update_types do
        types[next_num + i - 1] = update_types[i]
      end
      local query = format("update %s\nset %s\nwhere %s %s", self.quoted, updates, conditions, returning)
      return spi.prepare(query, unpack(types))
    end
    
    function pg_table_object_funcs:row_deleter( fields, returning )
      returning = fields_select(returning, '')
      if returning ~= '' then
        returning = 'returning '..returning
      end
      local conditions_sql = fields_equiv(fields)
      conditions_sql = concat(conditions_sql, "\n   AND ")
      local conditions_types = self:field_types(fields)
      local query = format("delete from %s\nwhere %s %s", self.quoted, conditions_sql, returning)
      return spi.prepare(query, unpack(conditions_types))
    end
end


function spi.run_initial_script(id, name, body, warn)
    local name = format("%d_%s", id, name)
    name = format("plluac.initial_scripts[%q]", name)
    local func, err = loadstring(body, name)
    if not func then
      warn(err)
    else
      local good, err = pcall(func)
      if not good then
        warn(err)
      end
    end
end

local funcs = spi.execute("select id, name, body from plluac.initial_scripts")
for _, row in ipairs(funcs or {}) do
    spi.run_initial_script(row.id, row.name, row.body, spi.warning)
end
