#include <pllua.h>

#ifdef PG_MODULE_MAGIC
PG_MODULE_MAGIC;
#endif

/* errmessage ought to be a const string with one %u inside */
/* caller should ReleaseSysCache(retval) when returned value is not nessesary */
HeapTuple SSearchSysCache(Oid kind, Oid value, const char *errmessage)
{
    HeapTuple Tup;
    Tup = SearchSysCache(kind, ObjectIdGetDatum(value), 0, 0, 0);
    if (!HeapTupleIsValid(Tup)) {
        elog(ERROR, errmessage, value);
    }
    return Tup;
}



/**********************************************************************
 * The information we cache about loaded procedures
 **********************************************************************/

#define PLLUA_PROC "PLLUA_PROC"

bool pllua_inited = false;
bool pllua_firstcall = true;
lua_State *LUA = NULL;
lua_State *LUA_TEMP = NULL;


/* these are saved and restored by pllua_call_handler */
pllua_call_info *current_call_info = NULL;

/**********************************************************************
 * Forward declarations
 **********************************************************************/
static void pllua_init_all(void);

Datum pllua_call_handler(PG_FUNCTION_ARGS);
Datum pllua_call_handler_simple(PG_FUNCTION_ARGS);
Datum pllua_validator(PG_FUNCTION_ARGS);
void _PG_init(void);

static Datum pllua_func_handler(PG_FUNCTION_ARGS);
static Datum pllua_trigger_handler(PG_FUNCTION_ARGS);
static void compile_pllua_function(Oid fn_oid, int is_trigger);

#include "boot.lua.h"
#include "init.lua.h"
const char PLLUA_BOOT[] = BOOT_LUA;
const char PLLUA_INIT[] = INIT_LUA;

/* Perform initialization during postmaster startup. */
static void compile_pllua_function(Oid fn_oid, int is_trigger);
static int luaP_compile_pllua_function(lua_State * L);

void _PG_init(void)
{
    int top = 0;

    if (pllua_inited)
        return;

    LUA = luaL_newstate();

    if (!LUA)
        elog(ERROR, "cannot create Lua state");
    luaL_openlibs(LUA);
    lua_settop(LUA, 0);
    lua_newtable(LUA);
    lua_setfield(LUA, LUA_REGISTRYINDEX, PLLUA_COMPILLED_FUNCTIONS);

    LUA_TEMP = lua_newthread(LUA);
    if (LUA_TEMP == NULL) {
        elog(ERROR, "cannot create temp lua state");
    }
    lua_setfield(LUA, LUA_REGISTRYINDEX, "PLLUA_TEMP_THREAD");

    luaP_register_conversations(LUA);
    luaP_registerspi(LUA);
    luaP_register_timestamp(LUA);

    /* NULL == NULL for simplicity. 
       Maybe it must be a userdata? or created on heap to be unique? */
    lua_pushlightuserdata(LUA, NULL);
    lua_setfield(LUA, LUA_GLOBALSINDEX, "NULL");

    /* now create functions and put then into registry in the reversed order */
    PLLUA_SPI_CONNECT();
    lua_settop(LUA, 0);
    if (luaL_loadbuffer(LUA, PLLUA_BOOT, BOOT_LUA_SIZE, "boot_lua"))
        elog(ERROR, "cannot load Lua startup script %s", lua_tostring(LUA, -1));
    lua_pushcfunction(LUA, luaP_compile_pllua_function);
    LUA_PCALL(1, 1, 0, "execution of startup script");
    lua_setfield(LUA, LUA_REGISTRYINDEX, PLLUA_MAKE_FUNCTION);
    PLLUA_SPI_FINISH();

    lua_gc(LUA, LUA_GCSETPAUSE, 220);
    lua_gc(LUA, LUA_GCSETSTEPMUL, 400);

    pllua_inited = true;
}

/* Perform initialization during backend startup. */
static void pllua_init_all(void)
{
    int top = lua_gettop(LUA);
    if (!pllua_inited)
        _PG_init();
    if (pllua_firstcall) {
        PLLUA_SPI_CONNECT();
        if (luaL_loadbuffer(LUA, PLLUA_INIT, INIT_LUA_SIZE, "init_lua"))
            elog(ERROR, "cannot load Lua init script %s",
                 lua_tostring(LUA, -1));
        LUA_PCALL(0, 0, 0, "execution of init script");
        PLLUA_SPI_FINISH();
        pllua_firstcall = false;
    }
}

/*
 * This is the first externally-visible part of the pllua call interface.
 * The Postgres function and trigger managers call it to execute a
 * lua function.
 */
//Datum pllua_call_handler_adv(PG_FUNCTION_ARGS, datum_to_lua_mode mode);
static Datum pllua_call_handler_adv(PG_FUNCTION_ARGS, datum_to_lua_mode mode)
{
    Datum retval;
    pllua_call_info *save_call_info;
    pllua_call_info local_call_info = PLLuaCallInfoInit;
    int top;
    lua_State *save_state;

#define restore() \
    do { \
        current_call_info = save_call_info; \
        LUA = save_state;       /* need this, cause SRF creates new state */ \
        lua_settop(LUA, top);   /* need this, cause actual handlers do not restore top */ \
    } while(0)

    pllua_init_all();

    top = lua_gettop(LUA);
    save_state = LUA;

    local_call_info.mode = mode;
    save_call_info = current_call_info;

    PG_TRY();
    {
        bool is_trigger = CALLED_AS_TRIGGER(fcinfo);

        /* Find or compile the function */
        compile_pllua_function(fcinfo->flinfo->fn_oid, is_trigger ? 1 : 0);

        local_call_info.prodesc = lua_touserdata(LUA, -2);;
        local_call_info.fcinfo = fcinfo;
        local_call_info.func_context = CurrentMemoryContext;
        local_call_info.func_resource = CurrentResourceOwner;
        local_call_info.fn_retistuple = local_call_info.prodesc->fn_retistuple;
        local_call_info.result_oid = local_call_info.prodesc->result_oid;

        current_call_info = &local_call_info;

        if (is_trigger)
            retval = PointerGetDatum(pllua_trigger_handler(fcinfo));
        else
            retval = pllua_func_handler(fcinfo);

    }
    PG_CATCH();
    {
        restore();
        lua_gc(LUA, LUA_GCCOLLECT, 0);
        PG_RE_THROW();
    }
    PG_END_TRY();

    restore();
    return retval;
}

PG_FUNCTION_INFO_V1(pllua_call_handler);
PG_FUNCTION_INFO_V1(pllua_call_handler_simple);

Datum pllua_call_handler(PG_FUNCTION_ARGS)
{
    return pllua_call_handler_adv(fcinfo, D2L_FULL);
}

Datum pllua_call_handler_simple(PG_FUNCTION_ARGS)
{
    return pllua_call_handler_adv(fcinfo, D2L_SIMPLE);
}

/*
 * This is the other externally visible function - it is called when CREATE
 * FUNCTION is issued to validate the function being created/replaced.
 */
PG_FUNCTION_INFO_V1(pllua_validator);

Datum pllua_validator(PG_FUNCTION_ARGS)
{
    Oid funcoid = PG_GETARG_OID(0);
    int top;

    pllua_init_all();

    top = lua_gettop(LUA);
    PG_TRY();
    {
        compile_pllua_function(funcoid, -1);
    }
    PG_CATCH();
    {
        lua_settop(LUA, top);
        PG_RE_THROW();
    }
    PG_END_TRY();

    /* the result of a validator is ignored */
    PG_RETURN_VOID();
}

/* searches the cache or compiles lua function and leaves on top of stack
   prodesc, function
*/
static int luaP_compile_pllua_function(lua_State * L)
{
    Oid fn_oid = luaL_checkinteger(L, 1);
    LUA_SPI_TRY();
    SPI_push();
    compile_pllua_function(fn_oid, -1);
    SPI_pop();
    LUA_SPI_CATCH();
    return 0;
}

static void compile_pllua_function(Oid fn_oid, int is_trigger)
{
    HeapTuple procTup;
    Form_pg_proc procStruct;
    pllua_proc_desc *prodesc = NULL;
    int i;
    int top;
    top = lua_gettop(LUA);

    /* We'll need the pg_proc tuple in any case... */
    procTup = SSearchSysCache(PROCOID, fn_oid,
                              "cache lookup failed for function %u");
    procStruct = (Form_pg_proc) GETSTRUCT(procTup);

    if (is_trigger != 1 && is_trigger != 0) {
        if (procStruct->prorettype == TRIGGEROID ||
            (procStruct->prorettype == OPAQUEOID && procStruct->pronargs == 0))
            is_trigger = 1;
        else
            is_trigger = 0;
    }

    /************************************************************
	 * Lookup the function oid in the hashtable
	 ************************************************************/

    lua_getfield(LUA, LUA_REGISTRYINDEX, PLLUA_COMPILLED_FUNCTIONS);
    lua_rawgeti(LUA, top + 1, fn_oid);

    if (lua_istable(LUA, -1)) {
        bool uptodate;

        lua_rawgeti(LUA, -1, 1);
        prodesc = (pllua_proc_desc *) lua_touserdata(LUA, -1);

                /************************************************************
		 * If it's present, must check whether it's still up to date.
		 * This is needed because CREATE OR REPLACE FUNCTION can modify the
		 * function's pg_proc entry without changing its OID.
		 ************************************************************/
        uptodate = (prodesc->fn_xmin == HeapTupleHeaderGetXmin(procTup->t_data)
#if PG_VERSION_NUM < 80300
                    && prodesc->fn_cmin ==
                    HeapTupleHeaderGetCmin(procTup->t_data)
#else
                    && ItemPointerEquals(&prodesc->fn_tid, &procTup->t_self)
#endif
            );

        if (!uptodate) {
            prodesc = NULL;
        } else {
            lua_rawgeti(LUA, -2, 2);
            lua_replace(LUA, top + 2);
            lua_replace(LUA, top + 1);
            ReleaseSysCache(procTup);
            return;
        }
        lua_pop(LUA, 1);
    }
    lua_pop(LUA, 1);


        /************************************************************
	 * If we haven't found it in the hashtable, we analyze
	 * the functions arguments and returntype and store
	 * the in-/out-functions in the prodesc block and create
	 * a new hashtable entry for it.
	 *
	 * Then we load the procedure into the Lua interpreter.
	 ************************************************************/
    if (prodesc == NULL) {
        HeapTuple langTup;
        Form_pg_language langStruct;
        Datum prosrcdatum;
        bool isnull;
        char *proc_source;

                /************************************************************
		 * Allocate a new procedure description block
		 ************************************************************/
        lua_createtable(LUA, 2, 0);
        prodesc = (pllua_proc_desc *)
            lua_newuserdata(LUA, sizeof(pllua_proc_desc) +
                            (!is_trigger ? procStruct->pronargs : 0) *
                            sizeof(Oid));
        if (prodesc == NULL)
            ereport(ERROR,
                    (errcode(ERRCODE_OUT_OF_MEMORY), errmsg("out of memory")));
        MemSet(prodesc, 0, sizeof(pllua_proc_desc));

        prodesc->fn_xmin = HeapTupleHeaderGetXmin(procTup->t_data);
#if PG_VERSION_NUM < 80300
        prodesc->fn_cmin = HeapTupleHeaderGetCmin(procTup->t_data);
#else
        prodesc->fn_tid = procTup->t_self;
#endif
        /* Remember if function is STABLE/IMMUTABLE */
        prodesc->fn_readonly =
            (procStruct->provolatile != PROVOLATILE_VOLATILE);

                /************************************************************
		 * Lookup the pg_language tuple by Oid
		 ************************************************************/
        langTup = SSearchSysCache(LANGOID, procStruct->prolang,
                                  "cache lookup failed for language %u");
        langStruct = (Form_pg_language) GETSTRUCT(langTup);
        prodesc->lanpltrusted = langStruct->lanpltrusted;
        ReleaseSysCache(langTup);

                /************************************************************
		 * Get the required information for input conversion of the
		 * return value.
		 ************************************************************/
        if (!is_trigger) {
            GET_TYPE_INFO(typeTup, typeStruct, procStruct->prorettype);

            /* Disallow pseudotype result, except VOID or RECORD */
            if (typeStruct->typtype == 'p') {
                if (procStruct->prorettype == VOIDOID ||
                    procStruct->prorettype == RECORDOID ||
                    procStruct->prorettype == ANYARRAYOID ||
                    procStruct->prorettype == ANYELEMENTOID)
                    /* okay */ ;
                else if (procStruct->prorettype == TRIGGEROID) {
                    ereport(ERROR,
                            (errcode(ERRCODE_FEATURE_NOT_SUPPORTED),
                             errmsg("trigger functions may only be called "
                                    "as triggers")));
                } else {
                    ereport(ERROR,
                            (errcode(ERRCODE_FEATURE_NOT_SUPPORTED),
                             errmsg("pllua functions cannot return type %s",
                                    format_type_be(procStruct->prorettype))));
                }
            }

            prodesc->result_oid = procStruct->prorettype;
            prodesc->fn_retisset = procStruct->proretset;
            prodesc->fn_retistuple = (typeStruct->typtype == 'c' ||
                                      procStruct->prorettype == RECORDOID);

            ReleaseSysCache(typeTup);
        }

        /************************************************************
		 * Get the required information for output conversion
		 * of all procedure arguments
		 ************************************************************/
        if (!is_trigger) {

            prodesc->nargs = procStruct->pronargs;

            for (i = 0; i < prodesc->nargs; i++) {
                GET_TYPE_INFO(typeTup, typeStruct,
                              procStruct->proargtypes.values[i]);

                /* Disallow pseudotype argument */
                if (typeStruct->typtype == 'p') {
                    if (procStruct->proargtypes.values[i] == ANYARRAYOID ||
                        procStruct->proargtypes.values[i] == ANYELEMENTOID)
                        /* okay */ ;
                    else
                        ereport(ERROR,
                                (errcode(ERRCODE_FEATURE_NOT_SUPPORTED),
                                 errmsg("pllua functions cannot take type %s",
                                        format_type_be(procStruct->proargtypes.
                                                       values[i]))));
                }

                prodesc->arg_oid[i] = procStruct->proargtypes.values[i];

                ReleaseSysCache(typeTup);
            }
        }

        /************************************************************
		 * Get the original function text.
		 ************************************************************/
        prosrcdatum = SysCacheGetAttr(PROCOID, procTup,
                                      Anum_pg_proc_prosrc, &isnull);
        if (isnull)
            elog(ERROR, "null prosrc");
        proc_source =
            DatumGetCString(DirectFunctionCall1(textout, prosrcdatum));

         /************************************************************
		 * Create the function in the interpreter
		 ************************************************************/
/*        if(prodesc->lanpltrusted)
        {
            lua_settop(LUA, top);
		    elog(ERROR, "Sorry, but I can't make trusted function so far");
        }
        else
*/
        {
            lua_getfield(LUA, LUA_REGISTRYINDEX, PLLUA_MAKE_FUNCTION);
        }
        /* there are empty table and prodesc on the top of stack */
        lua_pushstring(LUA, proc_source);
        lua_pushstring(LUA, NameStr(procStruct->proname));
        lua_pushinteger(LUA, fn_oid);

        if (!is_trigger) {
            if (prodesc->nargs == 0) {
                lua_pushstring(LUA, "");
            } else {
                bool isNull;
                Datum proargnames;
                proargnames = SysCacheGetAttr(PROCOID, procTup,
                                              Anum_pg_proc_proargnames,
                                              &isNull);
                if (isNull) {
                    if (!lua_checkstack(LUA, prodesc->nargs))
                        elog(FATAL, "cannot grow lua stack");
                    for (i = 0; i < prodesc->nargs; i++) {
                        lua_pushfstring(LUA, "_%d%s", i + 1,
                                        (i < prodesc->nargs - 1) ? "," : "");
                    }
                    lua_concat(LUA, prodesc->nargs);
                } else {
                    Datum *elems;
                    int nelems;
                    const char *nm;
                    int cnt = 0;
                    deconstruct_array(DatumGetArrayTypeP(proargnames),
                                      TEXTOID, -1, false, 'i', &elems, NULL,
                                      &nelems);
                    if (nelems != prodesc->nargs)       /* should not happen */
                        elog(ERROR, "internal error: proargnames is wrong");
                    if (!lua_checkstack(LUA, prodesc->nargs))
                        elog(FATAL, "cannot grow lua stack");
                    for (i = 0; i < nelems; i++) {
                        nm = DatumGetCString(DirectFunctionCall1
                                             (textout, elems[i]));
                        if (nm[0] == '\0') {
                            lua_pushfstring(LUA, "_%d%s", i + 1,
                                            (i < nelems - 1) ? "," : "");
                            cnt++;
                        } else
                            lua_pushfstring(LUA, "%s%s", nm,
                                            (i < nelems - 1) ? "," : "");
                    }
                    lua_concat(LUA, nelems);
                }
            }
        } else {
            lua_pushstring(LUA, "TD, NEW, OLD, ...");
        }
        PLLUA_SPI_CONNECT();

        LUA_PCALL(4, 1, 0, "compilation");

        PLLUA_SPI_FINISH();

        lua_rawseti(LUA, -3, 2);
        lua_rawseti(LUA, -2, 1);
        lua_rawseti(LUA, top + 1, fn_oid);

        pfree(proc_source);
    }

    lua_rawgeti(LUA, top + 1, fn_oid);
    lua_rawgeti(LUA, -1, 1);
    lua_rawgeti(LUA, -2, 2);
    lua_replace(LUA, top + 2);
    lua_replace(LUA, top + 1);

    ReleaseSysCache(procTup);
    return;
}


/* fills arguments of function */
static void
pllua_prepare_call_lua_func(pllua_proc_desc * desc, FunctionCallInfo fcinfo)
{
    int i;
    Oid actual;
    Fpush_datum push_datum;
    switch (current_call_info->mode) {
    case D2L_SIMPLE:
        push_datum = luaP_pushdatum_simple_cached;
        break;
    case D2L_STRINGS:
        push_datum = luaP_pushdatum_string_cached;
        break;
    case D2L_FULL:
    default:
        push_datum = luaP_pushdatum_cached;
        break;
    }

    for (i = 0; i < desc->nargs; i++) {
        if (fcinfo->argnull[i])
            lua_pushlightuserdata(LUA, NULL);
        else if ((desc->arg_oid[i] == ANYARRAYOID ||
                  desc->arg_oid[i] == ANYELEMENTOID) &&
                 (actual =
                  get_fn_expr_argtype(fcinfo->flinfo, i)) != InvalidOid)
            push_datum(LUA, fcinfo->arg[i], actual, NULL);
        else
            push_datum(LUA, fcinfo->arg[i], desc->arg_oid[i], NULL);

    }
}

static Datum pllua_func_handler(PG_FUNCTION_ARGS)
{
    pllua_proc_desc *prodesc = current_call_info->prodesc;
    Datum retval = (Datum) 0;
    ReturnSetInfo *rsi = NULL;
    //lua_State *saved_state = LUA;
    int top = lua_gettop(LUA);

    rsi = (ReturnSetInfo *) fcinfo->resultinfo;

    if (prodesc->fn_retisset) {
        /* Check context before allowing the call to go through */
        if (!rsi || !IsA(rsi, ReturnSetInfo) ||
            (rsi->allowedModes & SFRM_Materialize) == 0 ||
            rsi->expectedDesc == NULL) {
            ereport(ERROR,
                    (errcode(ERRCODE_FEATURE_NOT_SUPPORTED),
                     errmsg("set-valued function called in context that "
                            "cannot accept a set")));
        }
        lua_xmove(LUA, LUA_TEMP, 1);
        LUA = lua_newthread(LUA);
        lua_xmove(LUA_TEMP, LUA, 1);
    }

    if ((prodesc->result_oid == ANYARRAYOID
         || prodesc->result_oid == ANYELEMENTOID)
        && (current_call_info->result_oid =
            get_fn_expr_rettype(fcinfo->flinfo)) != InvalidOid) {
        GET_TYPE_INFO(heap, info, current_call_info->result_oid);
        current_call_info->fn_retistuple = (info->typtype == 'c');
        ReleaseSysCache(heap);
    } else {
        current_call_info->fn_retistuple = prodesc->fn_retistuple;
        current_call_info->result_oid = prodesc->result_oid;
    }

    if (current_call_info->fn_retistuple) {
        alloc_cache_type_heap_tupdesc(&current_call_info->cache_heap);
    }

    PLLUA_SPI_CONNECT();

    pllua_prepare_call_lua_func(prodesc, fcinfo);


    if (prodesc->fn_retisset) {
        int nargs = prodesc->nargs;
        int res_code;
        do {
            res_code = lua_resume(LUA, nargs);
            nargs = 0;
            switch (res_code) {
            case 0:
            case LUA_YIELD:{
                    int ytop = lua_gettop(LUA), i = 1;
                    lua_getfield(LUA, LUA_REGISTRYINDEX, "RETURN_NEXT");
                    for (i = 1; i <= ytop; i++) {
                        lua_pushvalue(LUA, i);
                    }
                    lua_call(LUA, ytop, 0);
                    break;
                }
                LUA_ERRORCASE("execution", {
                              }
                );
            }
        } while (res_code == LUA_YIELD);
        release_cache_type_heap(&current_call_info->cache_heap);
        rsi->returnMode = SFRM_Materialize;
        rsi->setResult = current_call_info->tuple_store;
        rsi->setDesc = current_call_info->cache_heap.tupdesc.tupdesc;

        retval = (Datum) 0;
        PLLUA_SPI_FINISH();
    } else {                    /* not SRF function */


        LUA_PCALL(prodesc->nargs, 1, 0, "execution");

        /************************************************************
        * Disconnect from SPI manager and then create the return
        * values datum (if the input function does a palloc for it
        * this must not be allocated in the SPI memory context
        * because SPI_finish would free it).
        ************************************************************/
        PLLUA_SPI_FINISH();

        if (lua_isNULL(LUA, -1) || current_call_info->result_oid == VOIDOID) {
            /* Return NULL if Lua code returned nil or NULL */
            if (rsi && IsA(rsi, ReturnSetInfo))
                rsi->isDone = ExprEndResult;
            if (current_call_info->result_oid != VOIDOID)
                fcinfo->isnull = true;
            retval = (Datum) 0;
        } else if (luaP_israwstring(LUA, -1)
                   || !current_call_info->fn_retistuple) {
            retval =
                luaP_todatum(LUA, current_call_info->result_oid, -1,
                             &fcinfo->isnull);
        } else {                // current_call_info->fn_retistuple
            /* Return a lua hash converted to a Datum */
            TupleDesc tupdesc = NULL;

            HeapTuple tup;

            if (!lua_istable(LUA, -1)) {
                ereport(ERROR,
                        (errcode(ERRCODE_DATATYPE_MISMATCH),
                         errmsg("composite-returning Lua function "
                                "must return table")));
            }

            /* XXX should cache the attinmeta data instead of recomputing */
            if (get_call_result_type(fcinfo, NULL, &tupdesc) !=
                TYPEFUNC_COMPOSITE) {
                ereport(ERROR,
                        (errcode(ERRCODE_FEATURE_NOT_SUPPORTED),
                         errmsg("function returning record called in context "
                                "that cannot accept type record")));
            }
            tupdesc = BlessTupleDesc(tupdesc);
            tup = luaP_form_tuple_with_tuple_desc(LUA, tupdesc);
            retval = HeapTupleGetDatum(tup);
        }
    }
    return retval;
}


/********************/
/* Trigger stuff    */
/********************/

/* Set up the arguments for a trigger call. */
static void pllua_trigger_build_args(FunctionCallInfo fcinfo)
{
    TriggerData *tdata;
    TupleDesc tupdesc;
    int i;
    int top = lua_gettop(LUA);
    char *level;
    char *event;
    char *when;
    char *name;

    tdata = (TriggerData *) fcinfo->context;
    tupdesc = tdata->tg_relation->rd_att;
    lua_createtable(LUA, 0, 7); /* TD */

#define SET(type, field, value) \
    do { lua_push##type(LUA, (value)); lua_setfield(LUA, top+1, (field));} while(0)

    SET(string, "NAME", tdata->tg_trigger->tgname);
    SET(integer, "TABLEID", tdata->tg_relation->rd_id);
    name = SPI_getnspname(tdata->tg_relation);
    SET(string, "SCHEMA", name);
    pfree(name);
    name = SPI_getrelname(tdata->tg_relation);
    SET(string, "TABLE", name);
    pfree(name);

    if (TRIGGER_FIRED_BY_INSERT(tdata->tg_event)) {
        event = "INSERT";
        if (TRIGGER_FIRED_FOR_ROW(tdata->tg_event)) {
            luaP_pushhash_adv_from_tuple(LUA, tupdesc, tdata->tg_trigtuple, current_call_info->mode);   /* NEW */
            lua_pushnil(LUA);   /* OLD */
        }
    } else if (TRIGGER_FIRED_BY_DELETE(tdata->tg_event)) {
        event = "DELETE";
        if (TRIGGER_FIRED_FOR_ROW(tdata->tg_event)) {
            lua_pushnil(LUA);   /* NEW */
            luaP_pushhash_adv_from_tuple(LUA, tupdesc, tdata->tg_trigtuple, current_call_info->mode);   /* OLD */
        }
    } else if (TRIGGER_FIRED_BY_UPDATE(tdata->tg_event)) {
        event = "UPDATE";
        if (TRIGGER_FIRED_FOR_ROW(tdata->tg_event)) {
            CacheTupleDesc(cache);
            cache.tupdesc = tupdesc;
            cache.do_not_release = true;
            luaP_pushhash_adv_from_tuple_with_cache_tuple_desc(LUA, &cache, tdata->tg_newtuple, current_call_info->mode);       /* NEW */
            luaP_pushhash_adv_from_tuple_with_cache_tuple_desc(LUA, &cache, tdata->tg_trigtuple, current_call_info->mode);      /* OLD */
            release_cache_tuple_desc(&cache);
        }
    } else
        event = "UNKNOWN";
    SET(string, "EVENT", event);

    if (TRIGGER_FIRED_FOR_ROW(tdata->tg_event))
        level = "ROW";
    else {
        lua_pushnil(LUA);
        lua_pushnil(LUA);
        if (TRIGGER_FIRED_FOR_STATEMENT(tdata->tg_event))
            level = "STATEMENT";
        else
            level = "UNKNOWN";
    }
    SET(string, "LEVEL", level);

    if (TRIGGER_FIRED_BEFORE(tdata->tg_event))
        when = "BEFORE";
    else if (TRIGGER_FIRED_AFTER(tdata->tg_event))
        when = "AFTER";
    else
        when = "UNKNOWN";
    SET(string, "WHEN", when);

    if (tdata->tg_trigger->tgnargs > 0)
        for (i = 0; i < tdata->tg_trigger->tgnargs; i++)
            lua_pushstring(LUA, tdata->tg_trigger->tgargs[i]);
#undef SET
}

typedef struct pllua_modify_tuple_closure {
    TupleDesc tupdesc;
    int *modattrs;
    Datum *modvalues;
    char *modnulls;
    Fto_datum to_datum;
    int slotsused;
} MTClosure, *PMTClosure;

static void push_field(PMTClosure closure)
{
    const char *key = lua_tostring(LUA, -2);
    int attn = SPI_fnumber(closure->tupdesc, key);
    Form_pg_attribute *attrs = closure->tupdesc->attrs;
    int slot = closure->slotsused;

    if (attn <= 0 || attrs[attn - 1]->attisdropped)
        ereport(ERROR,
                (errcode(ERRCODE_UNDEFINED_COLUMN),
                 errmsg("Lua table contains nonexistent column \"%s\"", key)));
    if (lua_isNULL(LUA, -1)) {
        closure->modvalues[slot] = (Datum) 0;
        closure->modnulls[slot] = 'n';
    } else {
        bool isnull = false;

        closure->modvalues[slot] = closure->to_datum(LUA,
                                                     attrs[attn -
                                                           1]->atttypid,
                                                     attrs[attn -
                                                           1]->
                                                     atttypmod, &isnull, NULL);
        closure->modnulls[slot] = isnull ? 'n' : ' ';
    }
    closure->modattrs[slot] = attn;
    closure->slotsused++;
}

static HeapTuple pllua_modify_tuple(TriggerData * tdata, HeapTuple otup)
{
    HeapTuple rtup;
    MTClosure closure;
    MemSet(&closure, 0, sizeof(closure));

    switch (current_call_info->mode) {
    case D2L_SIMPLE:
        closure.to_datum = luaP_todatum_simple_cached;
        break;
    case D2L_STRINGS:
        closure.to_datum = luaP_string_todatum_cached;
        break;
    case D2L_FULL:
    default:
        closure.to_datum = luaP_todatum_cached;
        break;
    }
    closure.tupdesc = tdata->tg_relation->rd_att;

    if (!lua_istable(LUA, -1))
        ereport(ERROR,
                (errcode(ERRCODE_DATATYPE_MISMATCH),
                 errmsg("TD.new is not a table")));

    closure.modattrs = palloc(closure.tupdesc->natts * sizeof(int));
    closure.modvalues = palloc(closure.tupdesc->natts * sizeof(Datum));
    closure.modnulls = palloc(closure.tupdesc->natts * sizeof(char));
    closure.slotsused = 0;

    lua_getfield(LUA, -1, "__mapping__");
    if (!lua_isnil(LUA, -1)) {
        lua_pushnil(LUA);
        while (lua_next(LUA, -2) != 0) {
            if (lua_type(LUA, -2) == LUA_TSTRING) {
                /* got field name and field position */
                lua_rawget(LUA, -4);    /* fetch field value */
                push_field(&closure);
            }
            lua_pop(LUA, 1);    /* removes `value'; keeps `key' for next iteration */
        }
        lua_pop(LUA, 1);        /* removes "__mapping__" or nil from stack */
        lua_pushnil(LUA);       /* first key for regular field trassert */
    }
    while (lua_next(LUA, -2) != 0) {
        /* `key' is at index -2 and `value' at index -1 */
        if (lua_type(LUA, -2) == LUA_TSTRING
            && strncmp(lua_tostring(LUA, -2), "__mapping__", 11) != 0) {
            push_field(&closure);
        }
        lua_pop(LUA, 1);        /* removes `value'; keeps `key' for next iteration */
    }

    if (closure.slotsused == 0) {
        rtup = otup;
    } else {
        rtup = SPI_modifytuple(tdata->tg_relation, otup,
                               closure.slotsused,
                               closure.modattrs,
                               closure.modvalues, closure.modnulls);
    }

    pfree(closure.modattrs);
    pfree(closure.modvalues);
    pfree(closure.modnulls);

    if (rtup == NULL)
        elog(ERROR, "SPI_modifytuple failed: %s",
             SPI_result_code_string(SPI_result));

    return rtup;
}


static Datum pllua_trigger_handler(PG_FUNCTION_ARGS)
{
    Datum retval;
    int top = lua_gettop(LUA);

    pllua_trigger_build_args(fcinfo);

    /* Connect to SPI manager */
    PLLUA_SPI_CONNECT();

    LUA_PCALL(lua_gettop(LUA) - top, 1, 0, "trigger execution");

    /************************************************************
	* Disconnect from SPI manager and then create the return
	* values datum (if the input function does a palloc for it
	* this must not be allocated in the SPI memory context
	* because SPI_finish would free it).
	************************************************************/
    PLLUA_SPI_FINISH();

    if (TRIGGER_FIRED_FOR_ROW(((TriggerData *) fcinfo->context)->tg_event)
        && TRIGGER_FIRED_BEFORE(((TriggerData *) fcinfo->context)->tg_event)) {
        if (lua_isnil(LUA, -1)) {
            /* nil result means go ahead with original tuple */
            TriggerData *trigdata = ((TriggerData *) fcinfo->context);

            if (TRIGGER_FIRED_BY_INSERT(trigdata->tg_event))
                retval = (Datum) trigdata->tg_trigtuple;
            if (TRIGGER_FIRED_BY_UPDATE(trigdata->tg_event))
                retval = (Datum) trigdata->tg_newtuple;
            else if (TRIGGER_FIRED_BY_INSERT(trigdata->tg_event)
                     || TRIGGER_FIRED_BY_DELETE(trigdata->tg_event))
                retval = (Datum) trigdata->tg_trigtuple;
            else
                retval = (Datum) 0;     /* can this happen? */
        } else if (lua_isNULL(LUA, -1) ||
                   (lua_type(LUA, -1) == LUA_TSTRING
                    && strncmp(lua_tostring(LUA, -1), "SKIP", 4) == 0)) {
            retval = PointerGetDatum(NULL);
        } else if (lua_istable(LUA, -1)) {
            HeapTuple trv;

            TriggerData *trigdata = (TriggerData *) fcinfo->context;
            if (TRIGGER_FIRED_BY_INSERT(trigdata->tg_event))
                trv = pllua_modify_tuple(trigdata, trigdata->tg_trigtuple);
            else if (TRIGGER_FIRED_BY_UPDATE(trigdata->tg_event))
                trv = pllua_modify_tuple(trigdata, trigdata->tg_newtuple);
            else {
                ereport(WARNING,
                        (errcode(ERRCODE_E_R_I_E_TRIGGER_PROTOCOL_VIOLATED),
                         errmsg("ignoring modified tuple in DELETE trigger")));
                trv = NULL;
            }
            retval = PointerGetDatum(trv);
        } else {
            ereport(ERROR,
                    (errcode(ERRCODE_E_R_I_E_TRIGGER_PROTOCOL_VIOLATED),
                     errmsg("result of Lua trigger function must be nil, "
                            "NULL or table with changes to apply")));
            return (Datum) 0;
        }
    } else {
        retval = PointerGetDatum(NULL);
    }

    return retval;
}
