#include "pllua.h"
#include "commands/portalcmds.h"
#include "commands/async.h"
#include "parser/parse_type.h"
//#include "utils/adt/builtins.h"

#define current_call_readonly \
  ( current_call_info ? current_call_info->prodesc->fn_readonly : true )
#define current_call_d2l_mode \
  ( current_call_info ? current_call_info->mode : D2L_FULL )

/**********************************************************************
 * The information we cache about prepared and saved plans
 **********************************************************************/
static void luaPP_parse_type_in(lua_State * L, int ind, Oid * type,
                                int32 * typmod, bool inlua);

typedef struct type_typemod {
    Oid type;
    int32 typmod;
    bool byval;
} type_typemod;

typedef struct luaP_query_desc {
    void *plan;
    int nargs;
    bool freed;
    type_typemod argtypes[1];
} luaP_query_desc;


static int luaP_push_processed_status(lua_State * L,
                                      SPITupleTable * tuptable, int processed,
                                      int status, datum_to_lua_mode d2l_mode);
static bool luaP_islimit(lua_State * L, int pos);
static int luaP_tolimit(lua_State * L, int pos);
static bool luaP_is_datum2lua_mode(lua_State * L, int pos);
static datum_to_lua_mode luaP_get_datum2lua_mode(lua_State * L, int pos);

static int luaP_execute(lua_State * L)
{
    /*
     * Execute the query inside a sub-transaction, so we can cope with errors
     * sanely
     */
    const char *query = luaL_checkstring(L, 1);
    int limit = 0, num = 0, top = lua_gettop(L);
    datum_to_lua_mode d2l_mode = current_call_d2l_mode;
    if (lua_gettop(L) > 1) {
        int i = 2;
        for (; i <= top; i++) {
            if (lua_type(L, i) == LUA_TNUMBER) {
                limit = lua_tointeger(L, i);
            } else if (luaP_islimit(L, i)) {
                limit = luaP_tolimit(L, i);
            } else if (luaP_is_datum2lua_mode(L, i)) {
                d2l_mode = luaP_get_datum2lua_mode(L, i);
            }
        }
    }

    LUA_SPI_TRY() {
        int spi_rv;
        spi_rv = SPI_execute(query, current_call_readonly, limit);
        num =
            luaP_push_processed_status(L, SPI_tuptable, SPI_processed, spi_rv,
                                       d2l_mode);
        if (SPI_tuptable) {
            luaP_clear_tuptable_index(L, SPI_tuptable->tupdesc);
            SPI_freetuptable(SPI_tuptable);
        }
    }
    LUA_SPI_CATCH();

    return num;
}

//void luaP_push_processed(lua_State *L, SPITupleTable *tuptable, int processed);
static void luaP_push_processed(lua_State * L, SPITupleTable * tuptable,
                                int processed, cache_tuple_desc * cache,
                                datum_to_lua_mode d2l_mode)
{
    int i;
    lua_createtable(L, processed, 0);
    for (i = 0; i < processed; ++i) {
        luaP_pushhash_adv_from_tuple_with_cache_tuple_desc(L, cache,
                                                           tuptable->vals[i],
                                                           d2l_mode);
        lua_rawseti(L, -2, i + 1);
    }
    luaL_getmetatable(L, PLLUA_RESULTSETMT);
    lua_setmetatable(L, -2);

}


static int luaP_push_processed_status(lua_State * L,
                                      SPITupleTable * tuptable, int processed,
                                      int status, datum_to_lua_mode d2l_mode)
{
    if (status > 0 && tuptable) {
        CacheTupleDesc(cache);
        cache.tupdesc = tuptable->tupdesc;
        cache.do_not_release = true;
        luaP_push_processed(L, tuptable, processed, &cache, d2l_mode);
        release_cache_tuple_desc(&cache);
    } else {
        lua_pushnil(L);
    }

    lua_pushstring(L, SPI_result_code_string(status));

    return 2;
}

static int luaP_query(lua_State * L)
{
    const char *query = luaL_checkstring(L, 1);
    /*
     * Execute the query inside a sub-transaction, so we can cope with errors
     * sanely
     */
    LUA_SPI_TRY() {
        void *plan;
        Portal portal;

        /* Create a cursor for the query */
        plan = SPI_prepare(query, 0, NULL);
        if (plan == NULL)
            elog(ERROR, "SPI_prepare() failed:%s",
                 SPI_result_code_string(SPI_result));

        portal = SPI_cursor_open(NULL, plan, NULL, NULL, false);
        SPI_freeplan(plan);
        if (portal == NULL)
            elog(ERROR, "SPI_cursor_open() failed:%s",
                 SPI_result_code_string(SPI_result));
        luaP_pushcursor(L, portal);
    }
    LUA_SPI_CATCH();

    return 1;
}

static int luaP_prepare(lua_State * L)
{
    const char *query = luaL_checkstring(L, 1);
    int top = lua_gettop(L);
    int i;
    luaP_query_desc *qdesc;
    Oid *types = NULL;
/*    for (i = 2; i <= top; i++) {
        luaL_checkstring(L, 1);
    }
*/
    qdesc =
        (luaP_query_desc *) lua_newuserdata(L,
                                            sizeof(luaP_query_desc) +
                                            sizeof(type_typemod) * (top - 1));
    if (top >= 2)
        types = (Oid *) palloc0(sizeof(Oid) * top - 1);
    qdesc->nargs = top - 1;
    qdesc->freed = false;

    LUA_SPI_TRY()
    void *plan;
    for (i = 2; i <= top; i++) {
        HeapTuple typeTup;
        Form_pg_type typeStruct;
        //const char *typename = lua_tostring(L, i);
        lua_pushvalue(L, i);
        luaPP_parse_type_in(L, top + 2, &qdesc->argtypes[i - 2].type,
                            &qdesc->argtypes[i - 2].typmod, false);
        lua_pop(L, 1);
        types[i - 2] = qdesc->argtypes[i - 2].type;
        typeTup =
            SearchSysCache(TYPEOID,
                           ObjectIdGetDatum(qdesc->argtypes[i - 2].type), 0, 0,
                           0);
        if (!HeapTupleIsValid(typeTup))
            elog(ERROR, "cache lookup failed for type %u",
                 qdesc->argtypes[i - 2].type);
        typeStruct = (Form_pg_type) GETSTRUCT(typeTup);
        qdesc->argtypes[i - 2].byval = typeStruct->typbyval;
        ReleaseSysCache(typeTup);
    }

                /************************************************************
		 * Prepare the plan and check for errors
		 ************************************************************/
    plan = SPI_prepare(query, qdesc->nargs, types);

    if (plan == NULL)
        elog(ERROR, "SPI_prepare() failed:%s",
             SPI_result_code_string(SPI_result));

                /************************************************************
		 * Save the plan into permanent memory (right now it's in the
		 * SPI procCxt, which will go away at function end).
		 ************************************************************/
    qdesc->plan = SPI_saveplan(plan);
    if (qdesc->plan == NULL)
        elog(ERROR, "SPI_saveplan() failed: %s",
             SPI_result_code_string(SPI_result));

    LUA_SPI_CATCH( {
                  if (types != NULL)
                  pfree(types);}
    );
    luaL_getmetatable(L, PLLUA_PLANMT);
    lua_setmetatable(L, -2);
    return 1;
}

bool luaP_checkmt(lua_State * L, int pos, const void *mt)
{
    if (!lua_isnone(L, pos) && lua_getmetatable(L, pos)) {
        const void *p = lua_topointer(L, -1);
        lua_pop(L, 1);
        return p == mt;
    }
    return false;
};

/* ======= ResultSet ======= */
static const void *luaP_resultsetmt_pointer = NULL;

bool luaP_isresultset(lua_State * L, int pos)
{
    return luaP_checkmt(L, pos, luaP_resultsetmt_pointer);
}

static int luaP_makeresultset(lua_State * L)
{
    int top = lua_gettop(L);
    if (top == 1) {
        if (!lua_istable(L, 1))
            luaL_typerror(L, 1, "table");
    } else {
        int i = 1;
        lua_createtable(L, top, 0);
        for (i = 1; i <= top; i++) {
            lua_pushvalue(L, i);
            lua_rawseti(L, -2, i);
        }
        lua_replace(L, 1);
        lua_settop(L, 1);
    }
    luaL_getmetatable(L, PLLUA_RESULTSETMT);
    lua_setmetatable(L, 1);
    return 1;
}

/* ======== Limit ======== */
/* limit is used only to distinguish from parameters in query:execute */
static const void *luaP_limitmt_pointer = NULL;

static bool luaP_islimit(lua_State * L, int pos)
{
    return luaP_checkmt(L, pos, luaP_limitmt_pointer);
}

static int luaP_tolimit(lua_State * L, int pos)
{
    int res;
    lua_rawgeti(L, pos, 1);
    res = lua_tointeger(L, -1);
    lua_pop(L, 1);
    return res;
}

static int luaP_limit(lua_State * L)
{
    luaL_checkint(L, 1);
    lua_createtable(L, 1, 0);
    lua_pushvalue(L, 1);
    lua_rawseti(L, -2, 1);
    luaL_getmetatable(L, PLLUA_LIMITMT);
    lua_setmetatable(L, -2);
    return 1;
}

static int luaP_limittoint(lua_State * L)
{
    lua_rawgeti(L, 1, 1);
    return 1;
}

/* ======= Datum2LuaMode ======== */
static const void *luaP_get_datum2lua_pointer = NULL;

static bool luaP_is_datum2lua_mode(lua_State * L, int pos)
{
    return luaP_checkmt(L, pos, luaP_get_datum2lua_pointer);
}

static void luaP_push_datum2lua_mode(lua_State * L, datum_to_lua_mode mode)
{
    datum_to_lua_mode *md =
        (datum_to_lua_mode *) lua_newuserdata(L, sizeof(datum_to_lua_mode));
    *md = mode;
    luaL_getmetatable(L, PLLUA_DATUM2LUASTRINGMT);
    lua_setmetatable(L, -2);
}

static datum_to_lua_mode luaP_get_datum2lua_mode(lua_State * L, int pos)
{
    datum_to_lua_mode *md = (datum_to_lua_mode *) lua_touserdata(L, pos);
    return *md;
}

static int luaP_d2lmode_current(lua_State * L)
{
    luaP_push_datum2lua_mode(L, current_call_d2l_mode);
    return 1;
}

static int luaP_d2lmode_within(lua_State * L)
{
    if (!current_call_info)
        luaL_error(L,
                   "could not change current d2l_mode during initialization");
    if (!lua_isfunction(L, 2))
        luaL_typerror(L, 2, "function");
    if (!luaP_is_datum2lua_mode(L, 1))
        luaL_typerror(L, 1, "d2l_mode");
    else {
        int top = lua_gettop(L);
        datum_to_lua_mode save_md = current_call_info->mode;
        int res;
        current_call_info->mode = luaP_get_datum2lua_mode(L, 1);
        res = lua_pcall(L, top - 2, LUA_MULTRET, 0);
        current_call_info->mode = save_md;
        switch (res) {
        case 0:
            return lua_gettop(L) - 1;
            break;
        case LUA_ERRRUN:
            lua_error(L);
            break;
        case LUA_ERRMEM:
            elog(PANIC, "pllua : memory error");
            break;
        default:
            elog(PANIC, "pllua : unknown error");
            break;
        }
    }
    return 0;
}

/* ======= Plan ========= */
static const void *luaP_planmt_pointer = NULL;

static bool luaP_isquerydesc(lua_State * L, int pos)
{
    return luaP_checkmt(L, pos, luaP_planmt_pointer);
}


static luaP_query_desc *luaP_toquerydesc(lua_State * L, int pos)
{
    luaP_query_desc *qdesc;
    if (!luaP_isquerydesc(L, pos))
        luaL_typerror(L, pos, "plan");
    qdesc = lua_touserdata(L, pos);
    if (qdesc->freed) {
        lua_pushstring(L, "plan freed");
        lua_error(L);
    }
    return qdesc;
}

static int luaP_query_free(lua_State * L)
{
    luaP_query_desc *qdesc;
    if (!luaP_isquerydesc(L, 1))
        luaL_typerror(L, 1, "plan");
    qdesc = lua_touserdata(L, 1);
    if (!qdesc->freed) {
        SPI_freeplan(qdesc->plan);
        qdesc->freed = true;
    }
    return 0;
}

static int luaP_query_freed(lua_State * L)
{
    luaP_query_desc *qdesc;
    if (!luaP_isquerydesc(L, 1))
        luaL_typerror(L, 1, "plan");
    qdesc = lua_touserdata(L, 1);
    lua_pushboolean(L, qdesc->freed);
    return 1;
}

static void luaP_fill_args(lua_State * L, Datum ** args, char **nulls,
                           luaP_query_desc * qdesc, int argstart)
{
    int i;
    bool isnull;
    *args = (Datum *) palloc0(sizeof(Datum) * qdesc->nargs);
    *nulls = (char *) palloc0(sizeof(char) * qdesc->nargs);
    for (i = 0; i < qdesc->nargs; i++) {
        lua_pushvalue(L, i + argstart);
        (*args)[i] =
            luaP_todatum(L, qdesc->argtypes[i].type, qdesc->argtypes[i].typmod,
                         &isnull);
        (*nulls)[i] = isnull ? 'n' : ' ';
        lua_pop(L, 1);
    }
}
static void luaP_free_args(Datum * args, luaP_query_desc * qdesc)
{
    int i;
    for (i = 0; i < qdesc->nargs; i++) {
        if (!qdesc->argtypes[i].byval && args[i] != PointerGetDatum(NULL)) {
            pfree(DatumGetPointer(args[i]));
            args[i] = PointerGetDatum(NULL);
        }
    }
    pfree(args);
}

static int luaP_queryexecute(lua_State * L)
{
    luaP_query_desc *qdesc = luaP_toquerydesc(L, 1);
    int top = lua_gettop(L);
    int argstart = 2;
    int limit = 0;
    datum_to_lua_mode d2l_mode = current_call_d2l_mode;
    int num, i;
    char *nulls = NULL;
    Datum *args = NULL;

    for (i = 2; i <= 3; i++) {
        if (luaP_islimit(L, i)) {
            limit = luaP_tolimit(L, i);
            argstart = i + 1;
        } else if (luaP_is_datum2lua_mode(L, i)) {
            d2l_mode = luaP_get_datum2lua_mode(L, i);
            argstart = i + 1;
        }
    }

    if (top != qdesc->nargs + argstart - 1) {
        luaL_error(L, "plan expected %d parameters, found %d", qdesc->nargs,
                   top - argstart + 1);
    }
    LUA_SPI_TRY()
    int spi_rv;
    if (qdesc->nargs > 0)
        luaP_fill_args(L, &args, &nulls, qdesc, argstart);
    spi_rv = SPI_execute_plan(qdesc->plan, args, nulls,
                              current_call_readonly, limit);
    num =
        luaP_push_processed_status(L, SPI_tuptable, SPI_processed, spi_rv,
                                   d2l_mode);
    if (SPI_tuptable) {
        luaP_clear_tuptable_index(L, SPI_tuptable->tupdesc);
        SPI_freetuptable(SPI_tuptable);
    }

    LUA_SPI_CATCH( {
                  if (nulls) pfree(nulls);
                  if (args) luaP_free_args(args, qdesc);}
    );
    return num;
}

static int luaP_queryquery(lua_State * L)
{
    luaP_query_desc *qdesc = luaP_toquerydesc(L, 1);
    int top = lua_gettop(L);
    int argstart = 2;
    char *nulls = NULL;
    Datum *args = NULL;

    if (top != qdesc->nargs + argstart - 1) {
        lua_pushfstring(L, "plan expected %d parameterd, found %d",
                        qdesc->nargs, top - argstart + 1);
        lua_error(L);
    }
    LUA_SPI_TRY()
    Portal portal;

    if (qdesc->nargs > 0)
        luaP_fill_args(L, &args, &nulls, qdesc, argstart);

    portal = SPI_cursor_open(NULL, qdesc->plan, args, nulls,
                             current_call_readonly);

    if (portal == NULL)
        elog(ERROR, "SPI_cursor_open() failed:%s",
             SPI_result_code_string(SPI_result));
    luaP_pushcursor(L, portal);

    LUA_SPI_CATCH( {
                  if (nulls) pfree(nulls);
                  if (args) luaP_free_args(args, qdesc);}
    );
    return 1;
}

/* ======= Cursor ======= */
static const void *luaP_cursormt_pointer = NULL;

bool luaP_iscursor(lua_State * L, int pos)
{
    return luaP_checkmt(L, pos, luaP_cursormt_pointer);
}

/*
  PortalCleanTupdesc - dirty hack. Maybe it doesn't reliable, but it just works
*/
static void PortalCleanTupdesc(Portal portal)
{
    luaP_clear_tuptable_index(LUA_TEMP, portal->tupDesc);
    portal->cleanup = PortalCleanup;
    PortalCleanup(portal);
}

void luaP_pushcursor(lua_State * L, Portal cursor)
{
    luaP_cursor *p = (luaP_cursor *) lua_newuserdata(L, sizeof(luaP_cursor));
    p->p = cursor;
    p->closed = false;
    cursor->cleanup = PortalCleanTupdesc;
    luaL_getmetatable(L, PLLUA_CURSORMT);
    lua_setmetatable(L, -2);
}

luaP_cursor *luaP_tocursor(lua_State * L, int pos)
{
    luaP_cursor *p;
    if (!luaP_iscursor(L, pos))
        luaL_typerror(L, pos, "cursor");
    p = (luaP_cursor *) lua_touserdata(L, pos);
    if (p->closed) {
        lua_pushstring(L, "cursor closed");
        lua_error(L);
    }
    if (!PortalIsValid(p->p)) {
        lua_pushstring(L, "cursor invalid");
        lua_error(L);
    }
    return p;
}

static int luaP_cursortostring(lua_State * L)
{
    luaP_cursor *p = luaP_tocursor(L, 1);
    lua_pushstring(L, p->p->name);
    return 1;
}

#if PG_VERSION_NUM < 80300

#define L_FETCH_DIRECTION_TYPE bool
#define FETCH_FORWARD true
#define FETCH_BACWARD false
#define SPI_SCROLL_FETCH SPI_cursor_fetch

#else

#define L_FETCH_DIRECTION_TYPE int
#define SPI_SCROLL_FETCH SPI_scroll_cursor_fetch

#endif
static L_FETCH_DIRECTION_TYPE luaP_find_direction(lua_State * L)
{
    L_FETCH_DIRECTION_TYPE dir = FETCH_FORWARD;
    int top = lua_gettop(L);
    int type = lua_type(L, 3);
    lua_getmetatable(L, 1);
    if (top <= 2 || type == LUA_TNIL) {
        dir = FETCH_FORWARD;
#if PG_VERSION_NUM >= 80300
    } else if (type == LUA_TNUMBER) {
        dir = lua_tointeger(L, 3);
        lua_rawgeti(L, -2, dir);
        if (lua_isnil(L, -1))
            dir = FETCH_FORWARD;
#else
    } else if (type == LUA_TBOOLEAN) {
        dir = lua_toboolean(L, 3);
#endif
    } else if (type == LUA_TSTRING) {
        lua_pushvalue(L, 3);
        lua_gettable(L, -2);
        if (!lua_isnil(L, -1))
#if PG_VERSION_NUM >= 80300
            dir = lua_tointeger(L, -1);
#else
            dir = lua_toboolean(L, -1);
#endif
    } else {
        dir = FETCH_FORWARD;
    }
    lua_settop(L, top);
    return dir;
}

static int luaP_cursorfetch(lua_State * L)
{
    luaP_cursor *cursor = luaP_tocursor(L, 1);
    long count = luaL_optlong(L, 2, FETCH_ALL);
    L_FETCH_DIRECTION_TYPE dir = luaP_find_direction(L);
    datum_to_lua_mode d2l_mode = current_call_d2l_mode;
    if (luaP_is_datum2lua_mode(L, 3))
        d2l_mode = luaP_get_datum2lua_mode(L, 3);
    else if (luaP_is_datum2lua_mode(L, 4))
        d2l_mode = luaP_get_datum2lua_mode(L, 4);
    SPI_SCROLL_FETCH(cursor->p, dir, count);
    if (SPI_processed > 0 && SPI_tuptable) {    /* any rows? */
        CacheTupleDesc(cache);
        cache.tupdesc = SPI_tuptable->tupdesc;
        cache.do_not_release = true;

        luaP_push_processed(L, SPI_tuptable, SPI_processed, &cache, d2l_mode);

        release_cache_tuple_desc(&cache);
    } else
        lua_pushnil(L);
    SPI_freetuptable(SPI_tuptable);
    return 1;
}

static int luaP_cursormove(lua_State * L)
{
    luaP_cursor *cursor = luaP_tocursor(L, 1);
    int count = luaL_optlong(L, 2, 1);
    L_FETCH_DIRECTION_TYPE dir = luaP_find_direction(L);
    SPI_SCROLL_FETCH(cursor->p, dir, count);
    return 0;
}

static int luaPP_cursor_iter_i(lua_State * L, datum_to_lua_mode d2l_mode)
{
    luaP_cursor *cursor = luaP_tocursor(L, 1);
    SPI_cursor_fetch(cursor->p, true, 1);
    if (SPI_processed > 0) {    /* any rows? */
        lua_pushboolean(L, true);
        luaP_pushhash_adv_from_tuple(L, SPI_tuptable->tupdesc,
                                     SPI_tuptable->vals[0], d2l_mode);
    } else {
        SPI_cursor_close(cursor->p);
        cursor->closed = true;
        lua_pushnil(L);
        lua_pushnil(L);
    }
    SPI_freetuptable(SPI_tuptable);
    return 2;
}

static int luaP_cursor_iter(lua_State * L)
{
    return luaPP_cursor_iter_i(L, current_call_d2l_mode);
}

static int luaP_cursor_iter_strings(lua_State * L)
{
    return luaPP_cursor_iter_i(L, D2L_STRINGS);
}

static int luaP_cursor_iter_simple(lua_State * L)
{
    return luaPP_cursor_iter_i(L, D2L_SIMPLE);
}

static int luaP_cursoriterate(lua_State * L)
{
    luaP_tocursor(L, 1);
    lua_getfield(L, 1, "__iter");
    lua_pushvalue(L, 1);
    lua_pushnil(L);
    return 3;
}

static int luaP_cursoriterate_strings(lua_State * L)
{
    luaP_tocursor(L, 1);
    lua_getfield(L, 1, "__iter_strings");
    lua_pushvalue(L, 1);
    lua_pushnil(L);
    return 3;
}

static int luaP_cursoriterate_simple(lua_State * L)
{
    luaP_tocursor(L, 1);
    lua_getfield(L, 1, "__iter_simple");
    lua_pushvalue(L, 1);
    lua_pushnil(L);
    return 3;
}

static int luaP_cursorclose(lua_State * L)
{
    if (!luaP_iscursor(L, 1)) {
        luaL_typerror(L, 1, "cursor");
    } else {
        luaP_cursor *cursor = (luaP_cursor *) lua_touserdata(L, 1);
        if (!cursor->closed && PortalIsValid(cursor->p)) {
            SPI_cursor_close(cursor->p);
            cursor->closed = true;
        }
    }
    return 0;
}

static int luaP_cursorclosed(lua_State * L)
{
    if (!luaP_iscursor(L, 1)) {
        luaL_typerror(L, 1, "cursor");
    } else {
        luaP_cursor *cursor = (luaP_cursor *) lua_touserdata(L, 1);
        lua_pushboolean(L, (cursor->closed || !PortalIsValid(cursor->p)));
        return 1;
    }
    return 0;
}

#if PG_VERSION_NUM >= 80300
static int luaP_cursorindex(lua_State * L)
{
    luaP_cursor *cursor = luaP_tocursor(L, 1);
    if (lua_isnumber(L, 2)) {
        SPI_scroll_cursor_fetch(cursor->p, FETCH_ABSOLUTE, lua_tointeger(L, 2));
        if (SPI_processed > 0) {        /* any rows? */
            datum_to_lua_mode d2l_mode = current_call_d2l_mode;
            if (luaP_is_datum2lua_mode(L, 2))
                d2l_mode = luaP_get_datum2lua_mode(L, 2);
            luaP_pushhash_adv_from_tuple(L, SPI_tuptable->tupdesc,
                                         SPI_tuptable->vals[0], d2l_mode);
        } else {
            lua_pushnil(L);
        }
        SPI_freetuptable(SPI_tuptable);
    } else {
        lua_getmetatable(L, 1);
        lua_getfield(L, -1, "__index__");
        lua_replace(L, 1);
        lua_pop(L, 1);
        lua_gettable(L, 1);
    }
    return 1;
}
#endif
/*********** rawstring *****************/
/* used when need to conver to datum
   string representation of value    **/
static const void *luaP_rawstringmt_pointer = NULL;

bool luaP_israwstring(lua_State * L, int pos)
{
    return luaP_checkmt(L, pos, luaP_rawstringmt_pointer);
}

static int luaP_rawstring(lua_State * L)
{
    if (!lua_isNULL(L, 1) && lua_type(L, 1) != LUA_TNUMBER) {
        lua_createtable(L, 1, 0);
        lua_pushvalue(L, 1);
        lua_rawseti(L, -2, 1);
        luaL_getmetatable(L, PLLUA_RAWSTRINGMT);
        lua_setmetatable(L, -2);
    }
    return 1;
}

/*********** spi *******************/
static int luaP_command_id(lua_State * L)
{
#if PG_VERSION_NUM < 80300
    lua_pushinteger(L, GetCurrentCommandId());
#else
    lua_pushinteger(L, GetCurrentCommandId(false));
#endif
    return 1;
}

static int luaP_findcursor(lua_State * L)
{
    const char *name = luaL_checkstring(L, 1);
    LUA_SPI_TRY()
    Portal cursor = SPI_cursor_find(name);
    luaP_pushcursor(L, cursor);
    LUA_SPI_CATCH();
    return 1;
}

static int luaP_info(lua_State * L)
{
    elog(INFO, "%s", lua_tostring(L, 1));
    return 0;
}

static int luaP_notice(lua_State * L)
{
    elog(NOTICE, "%s", lua_tostring(L, 1));
    return 0;
}

static int luaP_warning(lua_State * L)
{
    elog(WARNING, "%s", lua_tostring(L, 1));
    return 0;
}

static int luaP_fatal(lua_State * L)
{
    elog(FATAL, "%s", lua_tostring(L, 1));
    return 0;
}
static void luaPP_parse_type(lua_State * L, int ind, Oid * type, int32 * typmod)
{
    luaPP_parse_type_in(L, ind, type, typmod, true);
}

static void luaPP_parse_type_in(lua_State * L, int ind, Oid * type,
                                int32 * typmod, bool inlua)
{
    luaL_checkany(L, ind);
    switch (lua_type(L, ind)) {
    case LUA_TNUMBER:
        *type = lua_tointeger(L, ind);
        *typmod = luaL_optint(L, ind + 1, -1);
        break;
    case LUA_TSTRING:{
            const char *typename = lua_tostring(L, ind);
            if (inlua) {
                LUA_PG_TRY();
                parseTypeString(typename, type, typmod);
                LUA_PG_CATCH();
            } else {
                parseTypeString(typename, type, typmod);
            }
        }
        break;
    case LUA_TTABLE:
        lua_rawgeti(L, ind, 1);
        *type = lua_tointeger(L, -1);
        lua_rawgeti(L, ind, 2);
        *typmod = lua_isnil(L, -1) ? -1 : lua_tointeger(L, -1);
        lua_pop(L, 2);
        break;
    default:{
            const char *message = "typedefinition should be typename"
                " or typeoid (with typemod) or a table {typeoid, typemod}";
            if (inlua)
                luaL_error(L, message);
            else
                elog(ERROR, message);
        }
    }
}

static int luaP_parse(lua_State * L)
{
    Datum dat = 0;
    Oid type;
    int32 typmod;
    bool isnull;
    CacheTypeHeap(cache);

    luaPP_parse_type(L, 2, &type, &typmod);

    lua_pushvalue(L, 1);
    dat = luaP_string_todatum_cached(L, type, typmod, &isnull, &cache);
    lua_pop(L, 1);
    if (isnull)
        lua_pushNULL(L);
    else
        luaP_pushdatum_cached(L, dat, type, &cache);
    release_cache_type_heap(&cache);
    return 1;
}

static int luaP_tostring(lua_State * L)
{
    Datum dat = 0;
    Oid type;
    int32 typmod;
    bool isnull;
    CacheTypeHeap(cache);
    luaPP_parse_type(L, 2, &type, &typmod);

    lua_pushvalue(L, 1);
    dat = luaP_todatum_cached(L, type, typmod, &isnull, &cache);
    lua_pop(L, 1);
    if (isnull)
        lua_pushNULL(L);
    else
        luaP_pushdatum_string_cached(L, dat, type, &cache);
    release_cache_type_heap(&cache);
    return 1;
}

static int luaP_quote(lua_State * L)
{
    int i, nargs = lua_gettop(L);
    for (i = 1; i <= nargs; i++) {
        switch (lua_type(L, i)) {
        case LUA_TNIL:
            lua_pushstring(L, "NULL");
            break;
        case LUA_TNUMBER:
            lua_pushfstring(L, "%f", lua_tonumber(L, i));
            break;
        case LUA_TBOOLEAN:
            if (lua_toboolean(L, i)) {
                lua_pushstring(L, "true");
            } else {
                lua_pushstring(L, "false");
            }
            break;
        case LUA_TSTRING:{
                size_t llen = 0, len = 0;
                const char *s = luaL_checklstring(L, i, &llen), *cp1;
                char *res = (char *) palloc0(4 + llen * 2), *cp2;
                cp2 = res;
                len = llen;
                cp1 = s;
                for (; len-- > 0; cp1++) {
                    if (*cp1 == '\\') {
                        *cp2++ = ESCAPE_STRING_SYNTAX;
                        break;
                    }
                }
                *cp2++ = '\'';
                len = llen;
                cp1 = s;
                while (len-- > 0) {
                    if (SQL_STR_DOUBLE(*cp1, true))
                        *cp2++ = *cp1;
                    *cp2++ = *cp1++;
                }
                *cp2++ = '\'';
                lua_pushstring(L, res);
                pfree((void *) res);
            }
            break;
        default:
            luaL_error(L, "spi.qoute does not expects type %s",
                       lua_typename(L, i));
        }
        lua_replace(L, i);
    }
    return nargs;
}

static int luaP_quote_ident(lua_State * L)
{
    int i, nargs = lua_gettop(L);
    for (i = 1; i <= nargs; i++) {
        const char *in = luaL_checkstring(L, i);
        const char *res = quote_identifier(in);
        if (res != in) {
            lua_pushstring(L, res);
            pfree((void *) res);
            lua_replace(L, i);
        }
    }
    return nargs;
}

static int luaP_transaction(lua_State * L)
{
    int top = lua_gettop(L);
    if (!lua_isfunction(L, 1))
        luaL_error(L, "spi_transaction called with not function");
    do {
        MemoryContext oldcontext = CurrentMemoryContext;
        ResourceOwner oldowner = CurrentResourceOwner;

        BeginInternalSubTransaction(NULL);
        /* Want to run inside function's memory context */
        MemoryContextSwitchTo(oldcontext);
        switch (lua_pcall(L, top - 1, LUA_MULTRET, 0)) {
        case 0:{
                int i;
                ReleaseCurrentSubTransaction();
                MemoryContextSwitchTo(oldcontext);
                CurrentResourceOwner = oldowner;
                SPI_restore_connection();
                top = lua_gettop(L);
                lua_pushnil(L);
                for (i = top; i; i--) {
                    lua_pushvalue(L, i);
                    lua_replace(L, i + 1);
                }
                lua_pushboolean(L, true);
                lua_replace(L, 1);
                return top + 1;
            }
            break;
        case LUA_ERRRUN:{
                RollbackAndReleaseCurrentSubTransaction();
                MemoryContextSwitchTo(oldcontext);
                CurrentResourceOwner = oldowner;
                SPI_restore_connection();
                lua_pushvalue(L, 1);
                lua_pushboolean(L, false);
                lua_replace(L, 1);
                return 2;
            }
            break;
        case LUA_ERRMEM:
            elog(PANIC, "pllua : memory error");
            break;
        default:
            elog(PANIC, "pllua : unknown error");
            break;
        }
    } while (0);
    return 0;
}

static int luaP_notify(lua_State * L)
{
    const char *notify = luaL_checkstring(L, 1);
    SIMPLE_PG_TRY();
    Async_Notify(notify);
    SIMPLE_PG_CATCH();
    return 0;
}

static int luaP_listen(lua_State * L)
{
    const char *notify = luaL_checkstring(L, 1);
    SIMPLE_PG_TRY();
    Async_Listen(notify);
    SIMPLE_PG_CATCH();
    return 0;
}

static int luaP_unlisten(lua_State * L)
{
    const char *notify = luaL_checkstring(L, 1);
    SIMPLE_PG_TRY();
    Async_Unlisten(notify);
    SIMPLE_PG_CATCH();
    return 0;
}

#if PG_VERSION_NUM >= 80300
static int luaP_unlisten_all(lua_State * L)
{
    SIMPLE_PG_TRY();
    Async_UnlistenAll();
    SIMPLE_PG_CATCH();
    return 0;
}
#endif
/********** pg_type *************/

static int luaP_gettypeinfo(lua_State * L)
{
    const char *typename = luaL_checkstring(L, 2);
    Oid type;
    int32 typmod;
    LUA_PG_TRY();
    parseTypeString(typename, &type, &typmod);
    LUA_PG_CATCH();
    lua_createtable(L, 2, 0);
    lua_pushinteger(L, type);
    lua_rawseti(L, -2, 1);
    lua_pushinteger(L, typmod);
    lua_rawseti(L, -2, 2);
    return 1;
}

static int luaP_gettypename(lua_State * L)
{
    bool force = false;
    if (lua_isboolean(L, 2))
        force = lua_toboolean(L, 2);
    else if (lua_isboolean(L, 3))
        force = lua_toboolean(L, 2);
    if (lua_type(L, 1) == LUA_TSTRING && !force) {
        lua_settop(L, 1);
    } else {
        Oid type;
        int32 typmod = 0;
        Form_pg_type typ;
        HeapTuple typtup;

        luaPP_parse_type(L, 1, &type, &typmod);

        if (typmod != -1 || force) {
            char *typname = format_type_with_typemod(type, typmod);
            lua_pushstring(L, typname);
            pfree((void *) typname);
        } else {
            typtup = SSearchType(type);
            typ = (Form_pg_type) GETSTRUCT(typtup);
            lua_pushfstring(L, "%s", NameStr(typ->typname));
            ReleaseSysCache(typtup);
        }
    }
    return 1;
}
typedef struct DirectionConstantName {
    L_FETCH_DIRECTION_TYPE value;
    char *name;
} DirectionConstantName;

DirectionConstantName cursor_big_dirs[] = {
    {FETCH_FORWARD, "FORWARD"},
    {FETCH_BACKWARD, "BACKWARD"},
#if PG_VERSION_NUM >= 80300
    {FETCH_ABSOLUTE, "ABSOLUTE"},
    {FETCH_RELATIVE, "RELATIVE"},
#endif
    {0, NULL}
};
DirectionConstantName cursor_small_dirs[] = {
    {FETCH_FORWARD, "forward"},
    {FETCH_BACKWARD, "backward"},
#if PG_VERSION_NUM >= 80300
    {FETCH_ABSOLUTE, "absolute"},
    {FETCH_RELATIVE, "relative"},
#endif
    {0, NULL}
};

static void luaP_fill_direction_constants(lua_State * L,
                                          DirectionConstantName * nm)
{
    for (; nm->name; nm++) {
        lua_pushstring(L, nm->name);
#if PG_VERSION_NUM >= 80300
        lua_pushinteger(L, nm->value);
#else
        lua_pushboolean(L, nm->value);
#endif
        lua_pushvalue(L, -1);
        lua_pushvalue(L, -3);
        lua_settable(L, -5);
        lua_settable(L, -3);
    }
}


static const luaL_reg luaP_Cursor_funcs[] = {
    {"fetch", luaP_cursorfetch},
    {"move", luaP_cursormove},
    {"close", luaP_cursorclose},
    {"closed", luaP_cursorclosed},
    {"__iter", luaP_cursor_iter},
    {"__iter_strings", luaP_cursor_iter_strings},
    {"__iter_simple", luaP_cursor_iter_simple},
    {"iterate", luaP_cursoriterate},
    {"iterate_strings", luaP_cursoriterate_strings},
    {"iterate_simple", luaP_cursoriterate_simple},
#if PG_VERSION_NUM >= 80300
    {"get", luaP_cursorindex},
#endif
    {NULL, NULL}
};

static const luaL_reg luaP_Cursor_mt[] = {
    {"__tostring", luaP_cursortostring},
#if PG_VERSION_NUM >= 80300
    {"__index", luaP_cursorindex},
#endif
    {NULL, NULL}
};

static const luaL_reg luaP_Resultset_funcs[] = {
    //{"iterate", luaP_resultset_iterate},
    {NULL, NULL}
};

static const luaL_reg luaP_Limit_funcs[] = {
    {"value", luaP_limittoint},
    {NULL, NULL}
};

static const luaL_reg luaP_Query_funcs[] = {
    {"execute", luaP_queryexecute},
    {"query", luaP_queryquery},
    {"close", luaP_query_free},
    {"closed", luaP_query_freed},
    {NULL, NULL}
};

static const luaL_reg luaP_Query_mt[] = {
    {"__gc", luaP_query_free},
    {"__call", luaP_queryexecute},
    {NULL, NULL}
};

static const luaL_reg luaP_SPI_funcs[] = {
    {"execute", luaP_execute},
    {"find_cursor", luaP_findcursor},
    {"query", luaP_query},
    {"info", luaP_info},
    {"notice", luaP_notice},
    {"fatal", luaP_fatal},
    {"warning", luaP_warning},
    {"return_next", luaP_return_next},
    {"limit", luaP_limit},
    {"prepare", luaP_prepare},
    {"parse", luaP_parse},
    {"to_string", luaP_tostring},
    {"quote", luaP_quote},
    {"quote_ident", luaP_quote_ident},
    {"command_id", luaP_command_id},
    {"transaction", luaP_transaction},
    {"notify", luaP_notify},
    {"listen", luaP_listen},
    {"unlisten", luaP_unlisten},
#if PG_VERSION_NUM >= 80300
    {"unlisten_all", luaP_unlisten_all},
#endif
    {NULL, NULL}
};

static const luaL_reg luaP_String_funcs[] = {
    {"spi_parse", luaP_parse},
    {"spi_execute", luaP_execute},
    {"spi_query", luaP_query},
    {"spi_prepare", luaP_prepare},
    {"quote", luaP_quote},
    {"qoute_ident", luaP_quote_ident},
    {NULL, NULL}
};

void luaP_registerspi(lua_State * L)
{
    /* cursor */
    luaL_newmetatable(L, PLLUA_CURSORMT);
    luaL_register(L, "spi_cursor", luaP_Cursor_funcs);
    luaP_fill_direction_constants(L, cursor_big_dirs);
#if PG_VERSION_NUM >= 80300
    lua_setfield(L, -2, "__index__");
#else
    lua_setfield(L, -2, "__index");
#endif
    luaL_register(L, NULL, luaP_Cursor_mt);
    luaP_fill_direction_constants(L, cursor_big_dirs);
    luaP_fill_direction_constants(L, cursor_small_dirs);
    luaP_cursormt_pointer = lua_topointer(L, -1);
    lua_pop(L, 1);
    /* resultset */
    luaL_newmetatable(L, PLLUA_RESULTSETMT);
    luaL_register(L, "spi_resultset", luaP_Resultset_funcs);
    lua_setfield(L, -2, "__index");
    luaP_resultsetmt_pointer = lua_topointer(L, -1);
    lua_pop(L, 1);
    /* limit */
    luaL_newmetatable(L, PLLUA_LIMITMT);
    luaP_limitmt_pointer = lua_topointer(L, -1);
    lua_createtable(L, 0, 1);
    luaL_register(L, NULL, luaP_Limit_funcs);
    lua_setfield(L, -2, "__index");
    lua_pop(L, 1);
    /* query */
    luaL_newmetatable(L, PLLUA_PLANMT);
    luaP_planmt_pointer = lua_topointer(L, -1);
    luaL_register(L, NULL, luaP_Query_mt);
    lua_createtable(L, 0, 4);
    luaL_register(L, NULL, luaP_Query_funcs);
    lua_setfield(L, -2, "__index");
    lua_pop(L, 1);
    /* spi */
    luaL_register(L, "spi", luaP_SPI_funcs);
    lua_getfield(L, -1, "return_next");
    lua_setfield(L, LUA_REGISTRYINDEX, "RETURN_NEXT");
    lua_pop(L, 1);
    lua_pushcfunction(L, luaP_makeresultset);
    lua_setglobal(L, "set");
    /* rawparse */
    lua_pushcfunction(L, luaP_rawstring);
    lua_setglobal(L, "raw");
    luaL_newmetatable(L, PLLUA_RAWSTRINGMT);
    luaP_rawstringmt_pointer = lua_topointer(L, -1);
    lua_pop(L, 1);
    /* datum2lua_mode */
    lua_getglobal(L, "spi");
    luaL_newmetatable(L, PLLUA_DATUM2LUASTRINGMT);
    luaP_get_datum2lua_pointer = lua_topointer(L, -1);
    luaP_push_datum2lua_mode(L, D2L_FULL);
    lua_pushvalue(L, -1);
    lua_pushvalue(L, -1);
    lua_setfield(L, -4, "full");
    lua_setfield(L, -3, "Full");
    lua_setfield(L, -2, "FULL");
    luaP_push_datum2lua_mode(L, D2L_SIMPLE);
    lua_pushvalue(L, -1);
    lua_pushvalue(L, -1);
    lua_setfield(L, -4, "simple");
    lua_setfield(L, -3, "Simple");
    lua_setfield(L, -2, "SIMPLE");
    luaP_push_datum2lua_mode(L, D2L_STRINGS);
    lua_pushvalue(L, -1);
    lua_pushvalue(L, -1);
    lua_setfield(L, -4, "strings");
    lua_setfield(L, -3, "Strings");
    lua_setfield(L, -2, "STRINGS");
    lua_pushcfunction(L, luaP_d2lmode_current);
    lua_pushvalue(L, -1);
    lua_setfield(L, -3, "current");
    lua_setfield(L, -2, "Current");
    lua_createtable(L, 1, 0);
    lua_pushcfunction(L, luaP_d2lmode_within);
    lua_pushvalue(L, -1);
    lua_setfield(L, -3, "within");
    lua_setfield(L, -3, "within");
    lua_setfield(L, -1, "__index");
    lua_pushvalue(L, -1);
    lua_pushvalue(L, -1);
    lua_setfield(L, -2, "datum_to_lua_mode");
    lua_setfield(L, -3, "d2l_mode");
    lua_setfield(L, -2, "datum2lua_mode");
    lua_pop(L, 1);
    /* pg_type */
    lua_createtable(L, 0, 1);
    lua_pushcfunction(L, luaP_gettypename);
    lua_setfield(L, -2, "get_name");
    lua_createtable(L, 0, 1);
    lua_pushcfunction(L, luaP_gettypeinfo);
    lua_setfield(L, -2, "__index");
    lua_setmetatable(L, -2);
    lua_setglobal(L, "pg_type");
    /* string */
    luaL_register(L, "string", luaP_String_funcs);

};
