#include "pllua.h"
#include "commands/portalcmds.h"
#include "parser/parse_type.h"
#include "utils/timestamp.h"
#include "utils/datetime.h"
#include "utils/formatting.h"
#include <sys/time.h>

#ifdef HUGE_VAL
#define LUA_DATE_NOBEGIN	(-HUGE_VAL)
#define LUA_DATE_NOEND		(HUGE_VAL)
#else
#define LUA_DATE_NOBEGIN	(-DBL_MAX)
#define LUA_DATE_NOEND		(DBL_MAX)
#endif

#define SECS_SINCE_EPOCH ((int64)(POSTGRES_EPOCH_JDATE - UNIX_EPOCH_JDATE) * SECS_PER_DAY)
#define USECS_SINCE_EPOCH ((int64)(POSTGRES_EPOCH_JDATE - UNIX_EPOCH_JDATE) * SECS_PER_DAY * USECS_PER_SEC)

#if PG_VERSION_NUM <= 80300
#define session_timezone global_timezone
#endif
#ifndef USECS_PER_SEC
#define USECS_PER_SEC 1000000
#endif


static lua_Number timestamp_to_luanumber(TimestampTz ts)
{
    if (TIMESTAMP_IS_NOBEGIN(ts)) {
        return LUA_DATE_NOBEGIN;
    } else if (TIMESTAMP_IS_NOEND(ts)) {
        return LUA_DATE_NOEND;
    } else {
        lua_Number result;
#ifdef HAVE_INT64_TIMESTAMP
        result = ((lua_Number) ts) / USECS_PER_SEC;
#else
        result = (lua_Number) ts;
#endif
        return result;
    }
}

static TimestampTz luanumber_to_timestamp(lua_Number nm)
{
    if (LUA_DATE_NOBEGIN == nm) {
        return DT_NOBEGIN;
    } else if (LUA_DATE_NOEND == nm) {
        return DT_NOEND;
    } else {
        TimestampTz result;

#ifdef HAVE_INT64_TIMESTAMP
        result = (TimestampTz) rint(nm * USECS_PER_SEC);        // - USECS_SINCE_EPOCH;
#else
        result = (TimestampTz) nm;
#endif
        return result;
    }
}

int luaP_pushtimestamp(lua_State * L, TimestampTz ts)
{
    lua_pushnumber(L, timestamp_to_luanumber(ts));
    return 1;
}

TimestampTz luaP_totimestamp(lua_State * L, int pos)
{
    return luanumber_to_timestamp(lua_tonumber(L, pos));
}

static lua_Number luaP_timestamp2timestamptz(lua_State * L, lua_Number num)
{
    Timestamp timestamp = luanumber_to_timestamp(num);
    TimestampTz result;
    struct pg_tm tt, *tm = &tt;
    fsec_t fsec;
    int tz;

    if (TIMESTAMP_NOT_FINITE(timestamp))
        result = timestamp;
    else {
        if (timestamp2tm(timestamp, NULL, tm, &fsec, NULL, NULL) != 0)
            luaL_error(L, "timestamp.to_timestamptz: timestamp out of range");
        tz = DetermineTimeZoneOffset(tm, session_timezone);

        if (tm2timestamp(tm, fsec, &tz, &result) != 0)
            luaL_error(L, "timestamp.to_timestamptz: timestamp out of range");
    }

    return timestamp_to_luanumber(result);
}

static lua_Number luaP_timestamptz2timestamp(lua_State * L, lua_Number num)
{
    TimestampTz timestamp = luanumber_to_timestamp(num);
    Timestamp result;
    struct pg_tm tt, *tm = &tt;
    fsec_t fsec;
    char *tzn;
    int tz;

    if (TIMESTAMP_NOT_FINITE(timestamp))
        result = timestamp;
    else {
        if (timestamp2tm(timestamp, &tz, tm, &fsec, &tzn, NULL) != 0)
            luaL_error(L, "timestamptz.to_timestamp: timestamp out of range");
        if (tm2timestamp(tm, fsec, NULL, &result) != 0)
            luaL_error(L, "timestamptz.to_timestamp: timestamp out of range");
    }
    return timestamp_to_luanumber(result);
}

static int luaP_timestamp_to_timestamptz(lua_State * L)
{
    int i = 1, top = lua_gettop(L);
    for (i = 1; i <= top; i++) {
        lua_Number num = luaL_checknumber(L, i);
        lua_pushnumber(L, luaP_timestamp2timestamptz(L, num));
        lua_replace(L, i);
    }
    return top;
}

static int luaP_timestamptz_to_timestamp(lua_State * L)
{
    int i = 1, top = lua_gettop(L);
    for (i = 1; i <= top; i++) {
        lua_Number num = luaL_checknumber(L, i);
        lua_pushnumber(L, luaP_timestamptz2timestamp(L, num));
        lua_replace(L, i);
    }
    return top;
}

lua_Number luaP_get_clock(void)
{
    struct timeval tp;
    gettimeofday(&tp, NULL);
    return tp.tv_sec + (tp.tv_usec / 1000000.0);
}

lua_Number luaP_get_now(void)
{
    TimestampTz ts = GetCurrentTransactionStartTimestamp();
    return timestamp_to_luanumber(ts);
}

lua_Number luaP_get_statement_timestamp(void)
{
    TimestampTz ts = GetCurrentStatementStartTimestamp();
    return timestamp_to_luanumber(ts);
}

static int luaP_timestamptz_clock(lua_State * L)
{
    lua_pushnumber(L, luaP_get_clock());
    return 1;
}

static int luaP_timestamptz_now(lua_State * L)
{
    lua_pushnumber(L, luaP_get_now());
    return 1;
}

static int luaP_timestamptz_statement(lua_State * L)
{
    lua_pushnumber(L, luaP_get_statement_timestamp());
    return 1;
}

static int luaP_timestamp_to_char(lua_State * L)
{
    Timestamp ts = luanumber_to_timestamp(luaL_checknumber(L, 1));
    const char *fmts = luaL_checkstring(L, 2);
    Datum tsd = TimestampGetDatum(ts), fmt = string2text(fmts);
    Datum out = DirectFunctionCall2(timestamp_to_char, tsd, fmt);
    const char *outs = text2string(out);
    lua_pushstring(L, outs);
    pfree((void *) tsd);
    pfree((void *) fmt);
    pfree((void *) out);
    pfree((void *) outs);
    return 1;
}

static int luaP_timestamptz_to_char(lua_State * L)
{
    Timestamp ts = luanumber_to_timestamp(luaL_checknumber(L, 1));
    const char *fmts = luaL_checkstring(L, 2);
    Datum tsd = TimestampGetDatum(ts), fmt = string2text(fmts);
    Datum out = DirectFunctionCall2(timestamptz_to_char, tsd, fmt);
    const char *outs = text2string(out);
    lua_pushstring(L, outs);
    pfree((void *) tsd);
    pfree((void *) fmt);
    pfree((void *) out);
    pfree((void *) outs);
    return 1;
}

typedef int (*tz_func) (lua_State * L, Datum item, Datum tsd);
static int luaPP_timestampx_trunc_part(lua_State * L, tz_func tfunc)
{
    Timestamp ts = luanumber_to_timestamp(luaL_checknumber(L, 1));
    const char *fmts = luaL_checkstring(L, 2);
    Datum tsd = TimestampGetDatum(ts);
    Datum item = string2text(fmts);
    int nres = 0;
    LUA_PG_TRY();
    nres = tfunc(L, item, tsd);
    LUA_PG_CATCH((pfree((void *) tsd), pfree((void *) item)
                 ));
    return nres;
}

static int luaPP_timestamptz_trunc(lua_State * L, Datum item, Datum tsd)
{
    Datum out = DirectFunctionCall2(timestamptz_trunc, item, tsd);
    lua_Number num = timestamp_to_luanumber(DatumGetTimestampTz(out));;
    lua_pushnumber(L, num);
    pfree((void *) out);
    return 1;
}

static int luaPP_timestamp_trunc(lua_State * L, Datum item, Datum tsd)
{
    Datum out = DirectFunctionCall2(timestamp_trunc, item, tsd);
    lua_Number num = timestamp_to_luanumber(DatumGetTimestampTz(out));;
    lua_pushnumber(L, num);
    pfree((void *) out);
    return 1;
}

static int luaPP_timestamptz_part(lua_State * L, Datum item, Datum tsd)
{
    Datum out = DirectFunctionCall2(timestamptz_part, item, tsd);
    lua_Number num = DatumGetFloat8(out);
    lua_pushnumber(L, num);
    pfree((void *) out);
    return 1;
}

static int luaPP_timestamp_part(lua_State * L, Datum item, Datum tsd)
{
    Datum out = DirectFunctionCall2(timestamp_part, item, tsd);
    lua_Number num = DatumGetFloat8(out);
    lua_pushnumber(L, num);
    pfree((void *) out);
    return 1;
}

static int luaP_timestamptz_trunc(lua_State * L)
{
    return luaPP_timestampx_trunc_part(L, luaPP_timestamptz_trunc);
}

static int luaP_timestamp_trunc(lua_State * L)
{
    return luaPP_timestampx_trunc_part(L, luaPP_timestamp_trunc);
}

static int luaP_timestamptz_part(lua_State * L)
{
    return luaPP_timestampx_trunc_part(L, luaPP_timestamptz_part);
}

static int luaP_timestamp_part(lua_State * L)
{
    return luaPP_timestampx_trunc_part(L, luaPP_timestamp_part);
}

static int luaP_timestamptz_to_tm(lua_State * L)
{
    lua_Number num = luaL_checknumber(L, 1);
    if (num == LUA_DATE_NOBEGIN) {
        lua_pushstring(L, "-infinity");
    } else if (num == LUA_DATE_NOEND) {
        lua_pushstring(L, "infinity");
    } else {
        TimestampTz timestamp = luanumber_to_timestamp(num);
        struct pg_tm tt, *tm = &tt;
        fsec_t fsec;
        char *tzn;
        int tz;
        if (timestamp2tm(timestamp, &tz, tm, &fsec, &tzn, NULL) != 0)
            luaL_error(L, "timestamptz.to_timestamp: timestamp out of range");
        lua_createtable(L, 0, 0);
#define setfield(field) \
        lua_pushstring(L, #field); \
        lua_pushinteger(L, tm->tm_##field); \
        lua_rawset(L, -3);
        setfield(year);
        setfield(mon);
        setfield(mday);
        setfield(hour);
        setfield(min);
        setfield(sec);
        setfield(isdst);
        setfield(gmtoff);
#undef setfield
        lua_pushstring(L, "zone");
        lua_pushstring(L, tm->tm_zone);
        lua_rawset(L, -3);
        lua_pushstring(L, "fsec");
        lua_pushnumber(L, fsec);
        lua_rawset(L, -3);
        lua_pushstring(L, "tz");
        lua_pushinteger(L, tz);
        lua_rawset(L, -3);
    }
    return 1;
}

static int luaP_timestamp_to_tm(lua_State * L)
{
    lua_Number num = luaL_checknumber(L, 1);
    if (num == LUA_DATE_NOBEGIN) {
        lua_pushstring(L, "-infinity");
    } else if (num == LUA_DATE_NOEND) {
        lua_pushstring(L, "infinity");
    } else {
        TimestampTz timestamp = luanumber_to_timestamp(num);
        struct pg_tm tt, *tm = &tt;
        fsec_t fsec;
        if (timestamp2tm(timestamp, NULL, tm, &fsec, NULL, NULL) != 0)
            luaL_error(L, "timestamptz.to_timestamp: timestamp out of range");
        lua_createtable(L, 0, 0);
#define setfield(field) \
        lua_pushstring(L, #field); \
        lua_pushinteger(L, tm->tm_##field); \
        lua_rawset(L, -3);
        setfield(year);
        setfield(mon);
        setfield(mday);
        setfield(hour);
        setfield(min);
        setfield(sec);
        setfield(isdst);
        setfield(gmtoff);
#undef setfield
        lua_pushstring(L, "fsec");
        lua_pushnumber(L, fsec);
        lua_rawset(L, -3);
    }
    return 1;
}


static int luaP_timestamptz_from_tm(lua_State * L)
{
    if (lua_isstring(L, 1)) {
        const char *s = lua_tostring(L, 1);
        if (strncmp(s, "+infinity", 10) == 0 || strncmp(s, "infinity", 9) == 0) {
            lua_pushnumber(L, LUA_DATE_NOEND);
        } else if (strncmp(s, "-infinity", 10) == 0) {
            lua_pushnumber(L, LUA_DATE_NOBEGIN);
        } else {
            luaL_error(L, "timestampx.from_tm accepts table with fields "
                       "'year','mon','mday','hour','min','sec','tz'"
                       " and strings '+infinity','infinity','-infinity'\n"
                       " got string %s", s);
        }
    } else if (lua_istable(L, 1)) {
        struct pg_tm tt, *tm = &tt;
        fsec_t fsec;
        int tz, *tzp = &tz;
        TimestampTz result;
#define getfield(field, default) \
        lua_getfield(L, 1, #field); \
        tt.tm_##field = lua_isnil(L, -1) ? default : lua_tointeger(L, -1); \
        lua_pop(L, 1);
        getfield(year, 1970);
        getfield(mon, 1);
        getfield(mday, 1);
        getfield(hour, 0);
        getfield(min, 0);
        getfield(sec, 0);
#undef getfield
        lua_getfield(L, 1, "fsec");
        fsec = lua_isnil(L, -1) ? 0 : lua_tonumber(L, -1);
        lua_pop(L, 1);
        lua_getfield(L, 1, "tz");
        if (lua_isnil(L, -1)) {
            tzp = NULL;
        } else {
            tz = lua_tonumber(L, -1);
        }
        if (tm2timestamp(tm, fsec, tzp, &result) != 0)
            luaL_error(L, "timestamp.to_timestamptz: timestamp out of range");
        lua_pushnumber(L, timestamp_to_luanumber(result));
    } else {
        luaL_error(L, "timestampx.from_tm accepts table with fields "
                   "'year','mon','mday','hour','min','sec','tz'"
                   " and strings '+infinity','infinity','-infinity'");
    }
    return 1;
}


static int luaPP_timestamp_add_normalized_interval(lua_State * L,
                                                   PGFunction func)
{
    lua_Number num = luaL_checknumber(L, 1), secs = 0;
    TimestampTz tsz = luanumber_to_timestamp(num);
    Interval span = { 0, 0, 0 };
    if (!lua_istable(L, 2)) {
        luaL_error(L,
                   "timestampx.add_normalized_interval expects table as a second parameter");
    }
    lua_settop(L, 2);
    lua_getfield(L, 2, "seconds");
    if (!lua_isnil(L, -1)) {
        secs = lua_tonumber(L, -1);
#ifdef HAVE_INT64_TIMESTAMP
        span.time = (int64) rint(secs * USECS_PER_SEC);
#else
        span.time = secs;
#endif
    }
    lua_pop(L, 1);
    lua_getfield(L, 2, "days");
    if (!lua_isnil(L, -1)) {
        lua_Number days = lua_tonumber(L, -1);
        span.day = days;
    }
    lua_pop(L, 1);
    lua_getfield(L, 2, "months");
    if (!lua_isnil(L, -1)) {
        lua_Number months = lua_tonumber(L, -1);
        span.month = months;
    }
    lua_pop(L, 1);
    if (span.month != 0 || span.day != 0) {
        SIMPLE_PG_TRY();
        {
            Datum tszd = TimestampTzGetDatum(tsz);
            Datum res_ts =
                DirectFunctionCall2(func, tszd, IntervalPGetDatum(&span));
            tsz = DatumGetTimestampTz(res_ts);
            pfree((void *) tszd);
            pfree((void *) res_ts);
        }
        SIMPLE_PG_CATCH();
        lua_pushnumber(L, timestamp_to_luanumber(tsz));
    } else {
        lua_pushnumber(L, num + secs);
    }
    return 1;
}

static int luaP_timestamp_add_normalized_interval(lua_State * L)
{
    return luaPP_timestamp_add_normalized_interval(L, timestamp_pl_interval);
}

static int luaP_timestamptz_add_normalized_interval(lua_State * L)
{
    return luaPP_timestamp_add_normalized_interval(L, timestamptz_pl_interval);
}

void luaPP_interval_push(lua_State * L, Interval * span)
{
    lua_createtable(L, 0, 3);
#ifdef HAVE_INT64_TIMESTAMP
    lua_pushnumber(L, (lua_Number) span->time / USECS_PER_SEC);
#else
    lua_pushnumber(L, span->time);
#endif
    lua_setfield(L, -2, "seconds");
    lua_pushinteger(L, span->day);
    lua_setfield(L, -2, "days");
    lua_pushinteger(L, span->month);
    lua_setfield(L, -2, "months");
}

typedef struct FieldWeight {
    const char *name;
    double weight;
} FieldWeight;

void luaPP_interval_normalize(lua_State * L, int ind, Interval * span)
{
    const FieldWeight FWtimes[] = {
        {"seconds", 1},
        {"minutes", 60},
        {"hours", 3600},
        {NULL, 0}
    };
    const FieldWeight FWmonths[] = {
        {"months", 1},
        {"years", 12},
        {NULL, 0}
    };
    const FieldWeight *wtimes;
    span->time = 0;
    for (wtimes = FWtimes; wtimes->name != NULL; wtimes++) {
        lua_getfield(L, ind, wtimes->name);
        if (!lua_isnil(L, -1)) {
            lua_Number secs = lua_tonumber(L, -1) * wtimes->weight;
#ifdef HAVE_INT64_TIMESTAMP
            span->time += (int64) rint(secs * USECS_PER_SEC);
#else
            span->time += secs;
#endif
        }
        lua_pop(L, 1);
    }

    lua_getfield(L, ind, "days");
    if (!lua_isnil(L, -1)) {
        lua_Number days = lua_tonumber(L, -1);
        span->day = days;
    } else {
        span->day = 0;
    }
    lua_pop(L, 1);
    span->month = 0;
    for (wtimes = FWmonths; wtimes->name != NULL; wtimes++) {
        lua_getfield(L, ind, wtimes->name);
        if (!lua_isnil(L, -1)) {
            lua_Number months = lua_tonumber(L, -1) * wtimes->weight;
            span->month += months;
        }
        lua_pop(L, 1);
    }
}

static int luaP_interval_normalize(lua_State * L)
{
    Interval span = { 0, 0, 0 };
    luaPP_interval_normalize(L, 1, &span);
    luaPP_interval_push(L, &span);
    return 1;
}

static const luaL_reg luaP_Timestamp_funcs[] = {
    {"to_timestamptz", luaP_timestamp_to_timestamptz},
    {"to_char", luaP_timestamp_to_char},
    {"trunc", luaP_timestamp_trunc},
    {"part", luaP_timestamp_part},
    {"to_tm", luaP_timestamp_to_tm},
    {"from_tm", luaP_timestamptz_from_tm},
    {"add_normalized_interval", luaP_timestamp_add_normalized_interval},
    {NULL, NULL}
};

static const luaL_reg luaP_TimestampTz_funcs[] = {
    {"to_timestamp", luaP_timestamptz_to_timestamp},
    {"clock", luaP_timestamptz_clock},
    {"now", luaP_timestamptz_now},
    {"statement", luaP_timestamptz_statement},
    {"to_char", luaP_timestamptz_to_char},
    {"trunc", luaP_timestamptz_trunc},
    {"part", luaP_timestamptz_part},
    {"to_tm", luaP_timestamptz_to_tm},
    {"from_tm", luaP_timestamptz_from_tm},
    {"add_normalized_interval", luaP_timestamptz_add_normalized_interval},
    {NULL, NULL}
};
static const luaL_reg luaP_Interval_funcs[] = {
    {"normalize", luaP_interval_normalize},
    {NULL, NULL}
};

void luaP_register_timestamp(lua_State * L)
{
    lua_Number epoch_distance = SECS_SINCE_EPOCH;
    luaL_register(L, "timestamp", luaP_Timestamp_funcs);
    lua_pushnumber(L, -epoch_distance);
    lua_setfield(L, -2, "epoch");
    luaL_register(L, "timestamptz", luaP_TimestampTz_funcs);
    lua_pushnumber(L, -epoch_distance);
    lua_setfield(L, -2, "epoch");
    luaL_register(L, "interval", luaP_Interval_funcs);
    lua_pop(L, 3);
}
