/* Lua staff */
#include "lua.h"
#include "lauxlib.h"
#include "lualib.h"


#include "postgres.h"

/* system stuff */
#include <ctype.h>
#include <fcntl.h>
#include <unistd.h>

/* postgreSQL stuff */
#include "commands/trigger.h"
#include "executor/spi.h"
#include "funcapi.h"
#include "utils/lsyscache.h"
#include "utils/memutils.h"
#include "utils/typcache.h"
#include "miscadmin.h"
#include "mb/pg_wchar.h"
#include "catalog/pg_type.h"

/*
 * This routine is a crock, and so is everyplace that calls it.  The problem
 * is that the cached form of pllua functions/queries is allocated permanently
 * (mostly via malloc()) and never released until backend exit.  Subsidiary
 * data structures such as fmgr info records therefore must live forever
 * as well.  A better implementation would store all this stuff in a per-
 * function memory context that could be reclaimed at need.  In the meantime,
 * fmgr_info_cxt must be called specifying TopMemoryContext so that whatever
 * it might allocate, and whatever the eventual function might allocate using
 * fn_mcxt, will live forever too.
 */

#define perm_fmgr_info(finctionId, finfo) \
        fmgr_info_cxt((functionId),(finfo), TopMemoryContext)

extern lua_State *LUA;
extern lua_State *LUA_TEMP;


#define PLLUA_COMPILLED_FUNCTIONS "PLLUA_COMPILLED_FUNCTIONS"
#define PLLUA_TABLE_TO_PGARRAY    "PLLUA_TABLE_TO_PGARRAY"
#define PLLUA_MAKE_FUNCTION       "PLLUA_MAKE_FUNCTION"
#define PLLUA_CURSORMT "PLLUA_CURSOR"
#define PLLUA_RESULTSETMT "PLLUA_RESULTSET"
#define PLLUA_TUPTABLE_META "PLLUA_RECORD_META"
#define PLLUA_TUPTABLE_IDX "PLLUA_RECORD_INDEX"
#define PLLUA_PLANMT "PLLUA_PLANMT"
#define PLLUA_LIMITMT "PLLUA_LIMITMT"
#define PLLUA_RAWSTRINGMT "PLLUA_RAWSTRINGMT"
#define PLLUA_DATUM2LUASTRINGMT "PLLUA_DATUM2LUAMT"
/* errmessage ought to be a const string with one %u inside */
/* raises lua_error if SearchSysCahce returnes invalid HeapTuple */
/* caller should ReleaseSysCache(retval) when returned value is not nessesary */
HeapTuple SSearchSysCache(Oid kind, Oid value, const char *errmessage);

#define GET_TYPE_INFO(heap, info, typeoid) \
  HeapTuple heap = SSearchSysCache(TYPEOID, typeoid, "[pllua]: cache lookup failed for type %u"); \
  Form_pg_type info = (Form_pg_type) GETSTRUCT(heap)

#define SSearchType(type) \
   SSearchSysCache(TYPEOID, (type), "[pllua]: cache lookup failed for type %u")

#define lua_isNULL(L, i) \
  (lua_isnil(L, i) || (lua_islightuserdata(L, i) && lua_touserdata(L, i)==NULL ))
#define lua_pushNULL(L) \
  lua_pushlightuserdata(L, NULL)

/**********************************************/
/*          pllua-conversions.c               */
/**********************************************/
typedef enum datum_to_lua_mode {
  D2L_FULL,
  D2L_SIMPLE,
  D2L_STRINGS
} datum_to_lua_mode;

typedef struct cache_tuple_desc {
    TupleDesc tupdesc;
    bool do_not_release;
    struct cache_type_heap *cache_attr;
} cache_tuple_desc;             /* you should use MemSet(,0,sizeof(cache_tuple_desc)) to be good :-) */
#define CacheTupleDescInit {NULL, false, NULL}
#define CacheTupleDesc(cache) cache_tuple_desc cache = CacheTupleDescInit

typedef enum cache_type_heap_kind {
    HK_Simple = 1,
    HK_Tuple,
    HK_Array
} cache_type_heap_kind;

typedef struct cache_type_heap {
    HeapTuple heap;
    struct cache_type_heap *cache_elem;
    struct cache_tuple_desc tupdesc;
    FmgrInfo *finfo_in;
    FmgrInfo *finfo_out;
} cache_type_heap;
#define CacheTypeHeapInit {NULL, NULL, CacheTupleDescInit, NULL, NULL}
#define CacheTypeHeap(cache) cache_type_heap cache = CacheTypeHeapInit

int luaP_register_conversations(lua_State * L);
void luaP_clear_tuptable_index(lua_State * L, TupleDesc tupdesc);

void
luaP_pushhash_adv_from_tuple_with_cache_tuple_desc(lua_State * L,
                                               cache_tuple_desc * cache,
                                               HeapTuple tup,
                                               datum_to_lua_mode mode);
void luaP_pushhash_adv_from_tuple(lua_State * L, TupleDesc tupdesc, HeapTuple tup, datum_to_lua_mode mode);
#define luaP_pushhash_from_tuple_with_cache_tuple_desc(L, cache, tup) \
        luaP_pushhash_adv_from_tuple_with_cache_tuple_desc(L, cache, tup, D2L_FULL)
#define luaP_pushhash_from_tuple(L, tupdesc, tup) \
        luaP_pushhash_adv_from_tuple(L, tupdesc, tup, D2L_FULL)

typedef void (*Fpush_datum) (lua_State * L, Datum dat, Oid type,
                             cache_type_heap * cache);

void luaP_pushdatum_string_cached(lua_State * L, Datum dat, Oid type,
                                  cache_type_heap * cache);
void luaP_pushdatum_cached(lua_State * L, Datum dat, Oid type,
                           cache_type_heap * cache);
void luaP_pushdatum_simple_cached(lua_State * L, Datum dat, Oid type,
                                  cache_type_heap * cache);
#define luaP_pushdatum(L, dat, type) luaP_pushdatum_cached((L), (dat), (type), NULL)

typedef Datum (*Fto_datum) (lua_State * L, Oid type, int32 typmod,
                                 bool * isnull, cache_type_heap * cache);
HeapTuple luaP_form_tuple_with_tuple_desc(lua_State * L, TupleDesc tupdesc);
HeapTuple luaP_form_tuple_with_cache_tuple_desc(lua_State * L,
                                                cache_tuple_desc * tupdesc);
Datum luaP_string_todatum_cached(lua_State * L, Oid type, int32 typmod,
                                 bool * isnull, cache_type_heap * cache);
Datum luaP_todatum_cached(lua_State * L, Oid type, int32 typmod, bool * isnull,
                          cache_type_heap * cache);
Datum
    luaP_todatum_simple_cached(lua_State * L, Oid type, int32 typmod, bool * isnull,
                               cache_type_heap * cache);
#define luaP_todatum(L, type, len, isnull) luaP_todatum_cached((L), (type), (len), (isnull), NULL)

void release_cache_tuple_desc(cache_tuple_desc * cache);
void alloc_cache_tuple_attrs(cache_tuple_desc * cache);
void alloc_cache_type_heap_elem(cache_type_heap * cache);
void alloc_cache_type_heap_tupdesc(cache_type_heap * cache);
void alloc_cache_type_heap_finfo_in(cache_type_heap * cache);
void alloc_cache_type_heap_finfo_out(cache_type_heap * cache);
void release_cache_type_heap(cache_type_heap * cache);


/**********************************************/
/*          pllua-timestamp.c                 */
/**********************************************/

int luaP_pushtimestamp(lua_State * L, TimestampTz ts);
TimestampTz luaP_totimestamp(lua_State * L, int pos);
void luaP_register_timestamp(lua_State * L);
lua_Number luaP_get_clock(void);
lua_Number luaP_get_now(void);
lua_Number luaP_get_statement_timestamp(void);
void luaPP_interval_push(lua_State *L, Interval *span);
void luaPP_interval_normalize(lua_State *L, int ind, Interval *span);

/**********************************************/
/*          pllua-spi.c                       */
/**********************************************/
bool luaP_checkmt(lua_State * L, int pos, const void *mt);

typedef struct luaP_cursor {
    Portal p;
    bool closed;
} luaP_cursor;

void luaP_pushcursor(lua_State * L, Portal cursor);
bool luaP_iscursor(lua_State * L, int pos);
luaP_cursor *luaP_tocursor(lua_State * L, int pos);

bool luaP_isresultset(lua_State * L, int pos);
bool luaP_israwstring(lua_State * L, int pos);

int luaP_return_next(lua_State * L);
void luaP_registerspi(lua_State * L);

/********************************************/
/********* pllua-api ************************/
/********************************************/

typedef struct pllua_proc_desc {
    TransactionId fn_xmin;
#if PG_VERSION_NUM < 80300
    CommandId fn_cmin;
#else
    ItemPointerData  fn_tid;
#endif
    bool fn_readonly;
    bool lanpltrusted;
    bool fn_retisset;           /* true, if function returns set */
    bool fn_retistuple;         /* true, if function returns tuple */
    Oid result_oid;             /* Oid of result type */
    int nargs;
    Oid arg_oid[1];
} pllua_proc_desc;

typedef struct pllua_call_info {
    pllua_proc_desc *prodesc;
    FunctionCallInfo fcinfo;
    MemoryContext func_context;
    ResourceOwner func_resource;
    bool fn_retistuple;
    Oid result_oid;
    Tuplestorestate *tuple_store;
    cache_type_heap cache_heap;
    datum_to_lua_mode mode;
} pllua_call_info;

#define PLLuaCallInfoInit { NULL, NULL, NULL, NULL, false, 0, NULL, CacheTypeHeapInit, D2L_FULL }
extern pllua_call_info *current_call_info;


/* SPI_INTERFACE SHORTCUTS */
#define LUA_SPI_TRY() \
   do {/* \
	 * Execute the query inside a sub-transaction, so we can cope with errors \
	 * sanely \
	 */ \
	MemoryContext oldcontext = CurrentMemoryContext; \
	ResourceOwner oldowner = CurrentResourceOwner; \
    \
	PG_TRY(); {

#define LUA_SPI_CATCH_LONG(Finally, Release) \
    /* Commit the inner transaction, return to outer xact context */ \
		MemoryContextSwitchTo(oldcontext); \
		CurrentResourceOwner = oldowner; \
    \
		/* \
		 * AtEOSubXact_SPI() should not have popped any SPI context, but just \
		 * in case it did, make sure we remain connected. \
		 */ \
		SPI_restore_connection(); \
		{ Finally; } \
	} \
	PG_CATCH(); \
	{ \
                ErrorData  *edata; \
    \
                /* Save error info */ \
                MemoryContextSwitchTo(oldcontext); \
                edata = CopyErrorData(); \
                if (edata->elevel <= ERROR) { \
                  FlushErrorState(); \
                  CurrentResourceOwner = oldowner; \
                  /* \
                  * If AtEOSubXact_SPI() popped any SPI context of the subxact, it will \
                  * have left us in a disconnected state.  We need this hack to return \
                  * to connected state. \
                  */ \
                  SPI_restore_connection(); \
                  { Finally; } \
                  { Release; } \
                  /* Punt the error to Lua */ \
                  luaL_error(L, "SPI: %s", edata->message); \
                \
                } else { \
                  PG_RE_THROW(); \
                } \
		/* Can't get here, but keep compiler quiet */ \
		return 0; \
	} \
	PG_END_TRY(); \
    } while(0)

#define LUA_SPI_CATCH(Finally) LUA_SPI_CATCH_LONG(Finally, {})

/* Some EXCEPTION HANDLING shortcuts */
#define LUAL_PCALL( LUA, NumArg, NumRes, ErrFunc, Stage ) \
        do {  int res_code = lua_pcall((LUA), (NumArg), (NumRes), (ErrFunc) ); \
          switch (res_code) \
        { \
        case 0: break; \
        LUAL_ERRORCASE(LUA, Stage, lua_settop((LUA),top) ); \
        } } while(0)

#define LUA_PCALL(NumArg, NumRes, ErrFunc, Stage )  \
        LUAL_PCALL(LUA, NumArg, NumRes, ErrFunc, Stage )

#define LUAL_ERRORCASE(LUA, Stage, Resume) \
        case LUA_ERRRUN: { \
            const char *p; \
            if ((p = lua_tostring((LUA),-1))) { \
                if(strncmp(p, "SPI: ", 5)==0)  \
                  p += 5; \
            } else \
                p = "<error message is not a string>"; \
            { Resume; } \
            elog(ERROR, "pllua : runtime error during " Stage ": %s",p); \
            } \
            break; \
        case LUA_ERRMEM: \
            { Resume; } \
            elog(PANIC, "pllua : memory error during " Stage); \
            break; \
        default: \
            { Resume; } \
            elog(PANIC, "pllua : unknown error %d during " Stage, res_code); \
            break

#define LUA_ERRORCASE( Stage, Resume) LUAL_ERRORCASE(LUA, Stage, Resume)
 /*END*/
    /* LUA PG_TRY exception handling shortcuts */
#define LUA_PG_TRY() do {\
        ResourceOwner oldowner = CurrentResourceOwner; \
        CurrentResourceOwner = ResourceOwnerCreate(oldowner, "LuaPgTry"); \
        PG_TRY(); {
#define LUA_PG_CATCH(Release) \
                ResourceOwnerRelease(CurrentResourceOwner, RESOURCE_RELEASE_BEFORE_LOCKS, true, false); \
                ResourceOwnerDelete(CurrentResourceOwner); \
                CurrentResourceOwner = oldowner; \
        } \
	PG_CATCH(); \
        { \
                ErrorData  *edata; \
    \
                /* Save error info */ \
                edata = CopyErrorData(); \
                ResourceOwnerRelease(CurrentResourceOwner, RESOURCE_RELEASE_BEFORE_LOCKS, true, false); \
                ResourceOwnerDelete(CurrentResourceOwner); \
                CurrentResourceOwner = oldowner; \
                if (edata->elevel <= ERROR) { \
                  FlushErrorState(); \
          \
                  { Release; } \
                  /* Punt the error to Lua */ \
                  luaL_error(L, "SPI: %s", edata->message); \
                                \
                } else { \
                  PG_RE_THROW(); \
                } \
		/* Can't get here, but keep compiler quiet */ \
        } \
	PG_END_TRY(); \
} while(0)
#define SIMPLE_PG_TRY() \
    do { \
      PG_TRY(); {
#define SIMPLE_PG_CATCH()  } \
      PG_CATCH(); {     \
        ErrorData  *edata = CopyErrorData(); \
        if (edata->elevel <= ERROR) { \
          FlushErrorState(); \
          luaL_error(L, "SPI: %s", edata->message); \
        } else { \
          PG_RE_THROW(); \
        } \
      } \
      PG_END_TRY(); \
    } while(0)
/* end */

/* input */
#define text2string(d) DatumGetCString(DirectFunctionCall1(textout, (d)))
#define varchar2string(d) DatumGetCString(DirectFunctionCall1(varcharout, (d)))
#define char2string(d) DatumGetCString(DirectFunctionCall1(bpcharout, (d)))
/* output: later we need to copy the result to upper context */
#define string2varchar(s, l) \
  DirectFunctionCall3(varcharin, CStringGetDatum((s)), \
      ObjectIdGetDatum(VARCHAROID), Int32GetDatum(VARHDRSZ + (l)))
#define string2char(s, l) \
  DirectFunctionCall3(bpcharin, CStringGetDatum((s)), \
      ObjectIdGetDatum(BPCHAROID), Int32GetDatum(VARHDRSZ + (l)))
#define string2text(s) \
  DirectFunctionCall1(textin, CStringGetDatum((s)))

#define PLLUA_SPI_CONNECT() \
  do {\
    int con = SPI_connect(), line = __LINE__; \
    if (con != SPI_OK_CONNECT) \
      elog(ERROR, "could not connect to SPI manager (returns %d at %s:%d)", con, __FILE__ , line); \
    /*else \
        elog(NOTICE, "SPI_connect() returns %d at %s:%d", con, __FILE__ , line);*/ \
  } while(0)

#define PLLUA_SPI_FINISH() \
  do {\
    int fin = SPI_finish(), line = __LINE__; \
    if (fin != SPI_OK_FINISH) \
        elog(ERROR, "SPI_finish() failed (returns %d at %s:%d)", fin, __FILE__ , line);\
    /*else \
        elog(NOTICE, "SPI_finish() returns %d at %s:%d", fin, __FILE__ , line);*/ \
  } while(0)

